# CompuCanvas
# User Interface
# Written by Linda and Scott Wills
# (c) 2009-2011 Scott & Linda Wills

import os, tkFont, subprocess, shutil, tkMessageBox
from Tkinter        import *
from tkFileDialog   import askopenfilename, asksaveasfilename, askdirectory
from tkMessageBox   import showerror, showwarning, showinfo, askyesno
from copy           import copy
from urllib         import urlopen
from Objects        import *
from DigiSim        import *
from platform       import uname

######################### Todos ########################################
#
# get wire highlighting below everything.
# rethink properties for multiple selected objects
# add toggles/properties for wire objects?

######################### User Interface Modes #########################

class NormalMode() :
    def __init__(Self, Parent) :
        Self.Parent = Parent
        for Object in Parent.Selected :
            Object.Highlight(False)
        Parent.Selected = []
        Parent.Canvas.bind("<ButtonPress-1>", Self.Parent.Mark_Mouse)
        Parent.Canvas.bind("<Control-ButtonPress-1>", Self.Parent.Mark_Mouse)
        Parent.Canvas.bind("<B1-Motion>", Self.Parent.Update_DragBox)
        Parent.Canvas.bind("<Control-B1-Motion>", Self.Parent.Update_DragBox)
        Parent.Canvas.bind("<ButtonRelease-1>", Self.Parent.Select)
        Parent.Canvas.bind("<Control-ButtonRelease-1>", Self.Parent.Add_Select)
        if Parent.OS == "Darwin" :
            Parent.Canvas.bind("<Control-Button-2>", Self.Parent.Toggle_Source)  # OSX
            Parent.Canvas.bind("<Button-2>", Self.Parent.Properties)
        else :
            Parent.Canvas.bind("<Control-Button-3>", Self.Parent.Toggle_Source)  # Linux, Windows
            Parent.Canvas.bind("<Button-3>", Self.Parent.Properties)

    def Update(Self, SelectedButton, ObjectType) :
        """ This routine handles a proto button push in normal mode.
        The current mode is exited and either a new place mode or a
        new connect mode is returned, depending on the button. """
        Self.Exit()
        if ObjectType == WIRE :
            return ConnectMode(Self.Parent, SelectedButton, ObjectType)
        else :
            return PlaceMode(Self.Parent, SelectedButton, ObjectType)

    def Exit(Self) :
        """ This routine handles leaving normal mode. """

class PlaceMode(NormalMode) :
    def __init__(Self, Parent, SelectedButton, ObjectType) :
        NormalMode.__init__(Self, Parent)
        Self.SelectedButton = SelectedButton
        Self.ObjectType = ObjectType
        Self.SelectedButton["background"] = "red"
        Parent.Canvas.bind("<Button-1>", Self.Parent.Place)

    def Update(Self, SelectedButton, ObjectType) :
        """ This routine handles a proto button push in place mode.
        If the current selected button is pressed, place mode is
        exited and a new normal mode is returned. If the wire
        button is pressed, the current mode is exited and a new
        connect mode is returned. Otherwise the selected object
        and type are updated. """
        Self.Exit()
        if SelectedButton == Self.SelectedButton :
             return NormalMode(Self.Parent)
        elif ObjectType == WIRE :
            return ConnectMode(Self.Parent, SelectedButton, ObjectType)
        else :
            Self.__init__(Self.Parent, SelectedButton, ObjectType)
            return Self

    def Exit(Self) :
        """ This routine handles leaving place mode. """
        Self.SelectedButton["background"] = "gray85"

class ConnectMode(NormalMode) :
    def __init__(Self, Parent, SelectedButton, ObjectType) :
        NormalMode.__init__(Self, Parent)
        Self.SelectedButton = SelectedButton
        Self.ObjectType = ObjectType
        Self.SelectedButton["background"] = "red"
        Parent.Canvas.Draw_Contacts()
        Parent.Canvas.bind("<ButtonRelease-1>", Self.Parent.Connect)     # HorzVert lock
        Parent.Canvas.bind("<Control-Button-1>", Self.Parent.Connect)    # unlocked
        if Parent.OS == "Darwin" :
            Parent.Canvas.bind("<Button-2>", Self.Parent.Remove_Point)   # OSX
        else :
            Parent.Canvas.bind("<Button-3>", Self.Parent.Remove_Point)   # Linux, Windows
        Self.NewWire = None

    def Update(Self, SelectedButton, ObjectType) :
        """ This routine handles a proto button push in connect
        mode. If the wire button is pressed, connect mode is
        exited and a new normal mode is returned. If another
        proto button is pressed, connect mode is exited and a
        new place mode is returned. """
        Self.Exit()
        if ObjectType == WIRE :
            return NormalMode(Self.Parent)
        else:
            return PlaceMode(Self.Parent, SelectedButton, ObjectType)

    def Exit(Self) :
        """ This routine handles leaving connect mode. """
        Self.SelectedButton["background"] = "gray85"
        Self.Parent.Canvas.Erase_Contacts()
        if Self.NewWire :
            Self.Parent.Canvas.delete(Self.NewWire.ID)

######################### CompuCanvas Window #########################

class CompuCanvas(Tk) :
    def __init__(Self, Version) :
        Tk.__init__(Self)
        Self.Sim = Simulation(Self)
        Self.Version = Version
        Self.title("CompuCanvas %1.2f" % Version)

        # CompuCanvas environment variables
        Self.Objects = []
        Self.Selected = []
        Self.MouseX = None
        Self.MouseY = None
        Self.Clipboard = []
        Self.Width = 600
        Self.Height = 400
        Self.Scale = 2
        Self.LineWidth = 3
        Self.FontSize = "auto"
        Self.ButtonMode = "icon+name"
        Self.Grid = False
        Self.CurPaths = {"sch": None, "eps": None, "plt": None}
        Self.FileName = None
        Self.Modified = False
        Self.Comment = None
        Self.Run = True
        Self.PropMod = False
        Self.SimDelay = 1000
        Self.Cycle = 0
        Self.PlotLength = 100
        Self.PlotWin = None
        Self.States = []
        Self.SwitchMarker = True
        Self.ColorMode = "color"
        Self.OS = (uname())[0]

        # load options parameters if options init file exists
        Self.Load_Init_File()
        if Self.Width < 400 :
            Self.Width = 400
        if Self.Height < 300 :
            Self.Height = 300

        # make canvas workspace
        Width = Self.Width*Self.Scale
        Height = Self.Height*Self.Scale
        Self.XSbar = Scrollbar(Self, orient=HORIZONTAL)
        Self.YSbar = Scrollbar(Self, orient=VERTICAL)
        Self.Canvas = CanvasView(Self, width=Width, height=Height, scrollregion=(0, 0, Width, Height),
                                 bg="white", xscrollcommand=Self.XSbar.set, yscrollcommand=Self.YSbar.set)
        Self.XSbar.config(command=Self.Canvas.xview)
        Self.YSbar.config(command=Self.Canvas.yview)

        # make button menu
        Self.Menu = MenuPane(Self)

        # make prototype display
        Self.Proto = ProtoPane(Self)

        # make starting mode
        Self.Mode = NormalMode(Self)

        # set grid geometry
        Self.Menu.grid(row=0, column=0, columnspan=3, sticky=N+S+E+W)
        Self.Proto.grid(row=1, column=0, rowspan=2, sticky=N)
        Self.Canvas.grid(row=1, column=1)
        Self.XSbar.grid(row=2, column=1, sticky=E+W)
        Self.YSbar.grid(row=1, column=2, sticky=N+S)

        # set grid resizing attributes
        Self.rowconfigure(1, weight=1)
        Self.columnconfigure(1, weight=1)

        # set shortcut keys
        Self.bind_class(Self, "<Control-l>", Self.Load)
        Self.bind_class(Self, "<Control-s>", Self.Save)
        Self.bind_class(Self, "<Control-S>", Self.SaveAs)
        Self.bind_class(Self, "<Control-x>", Self.Cut)
        Self.bind_class(Self, "<Control-c>", Self.Copy)
        Self.bind_class(Self, "<Control-v>", Self.Paste)
        Self.bind_class(Self, "<Control-o>", Self.Options)
        Self.bind_class(Self, "<Control-a>", Self.Select_All)
        Self.bind_class(Self, "<Control-g>", Self.Toggle_Grid)
        Self.bind_class(Self, "<Control-t>", Self.Tick)
        Self.bind_class(Self, "<Alt-c>", Self.Edit_Comment)
        Self.bind_class(Self, "<Alt-s>", Self.Write_EPS)
        Self.bind_class(Self, "<Alt-p>", Self.Write_Plot_EPS)
        Self.bind_class(Self, "<Alt-t>", Self.Trim)
        Self.bind_class(Self, "<Alt-g>", Self.Grow)     # percent grow
        Self.bind_class(Self, "<Alt-G>", Self.Grow)     # fixed small grow
        Self.bind_class(Self, "<Alt-P>", Self.Create_PNG)
        Self.bind_class(Self, "<Alt-D>", Self.Dir_Create_PNG)
        Self.bind_class(Self, "<Control-Up>", Self.Move_Up)
        Self.bind_class(Self, "<Control-Down>", Self.Move_Down)
        Self.bind_class(Self, "<Control-Left>", Self.Move_Left)
        Self.bind_class(Self, "<Control-Right>", Self.Move_Right)
        Self.bind_class(Self, "<Delete>", Self.Delete)
        Self.bind_class(Self, "<Control-Delete>", Self.Delete)
        Self.bind_class(Self, "<Control-q>", Self.Exit)

        # create top menubar
        Self.menubar = Menu(Self, relief=RAISED)
        Self.TopFileMenu() 
        Self.TopEditMenu()
        Self.TopHelpMenu()
        Self.config(menu=Self.menubar)

        # look for file to load in command line
        if len(sys.argv) > 1 :
            FilePath = sys.argv[1]
            Ext = FilePath.split('.')[-1]
            if Ext in ["sch", "SCH"] :
                Self.Load(FilePath=FilePath)        # load command line file
            else :
                showerror("Compucanvas", "Cannot open file type " + Ext)

        # start simulation events
        Self.after(1000, Self.SimLoop)

    def TopFileMenu(Self) :
        filemenu = Menu(Self.menubar, tearoff=0, relief=RAISED)
        Self.menubar.add_cascade(label="File", menu=filemenu)
        filemenu.add_command(label="New", underline=0, accelerator = "Ctrl-N", command=Self.New)
        filemenu.add_command(label="Load", underline=0, accelerator = "Ctrl-L", command=Self.Load)
        filemenu.add_command(label="Save", underline=0, accelerator = "Ctrl-S", command=Self.Save)
        filemenu.add_command(label="Save As", underline=5, accelerator = "Ctrl-Shift-S", command=Self.SaveAs)
        filemenu.add_command(label="Comment", underline=0, accelerator = "Alt-C", command=Self.Edit_Comment)
        filemenu.add_command(label="Write EPS", underline=8, accelerator = "Alt-S", command=Self.Write_EPS)
        filemenu.add_command(label="Write Plot EPS", underline=6, accelerator = "Alt-P", command=Self.Write_Plot_EPS)
        filemenu.add_separator()
        filemenu.add_command(label="Exit", accelerator = "Ctrl-Q", command=Self.quit)
        
    def TopEditMenu(Self) :
        editmenu = Menu(Self.menubar, tearoff=0, relief=RAISED)
        editmenu.add_command(label="Cut", underline=1, accelerator = "ctrl-X", command=Self.Cut)
        editmenu.add_command(label="Copy", underline=0, accelerator = "ctrl-C", command=Self.Copy)
        editmenu.add_command(label="Paste", underline=7, accelerator = "ctrl-X", command=Self.Paste)
        editmenu.add_command(label="Select All", underline=7, accelerator = "ctrl-A", command=Self.Select_All)
        editmenu.add_separator()
        editmenu.add_command(label="Move Up", accelerator = "ctrl-up", command=Self.Move_Up)
        editmenu.add_command(label="Move Down", accelerator = "ctrl-down", command=Self.Move_Down)
        editmenu.add_command(label="Move Left", accelerator = "ctrl-left", command=Self.Move_Left)
        editmenu.add_command(label="Move Right", accelerator = "ctrl-right", command=Self.Move_Right)
        editmenu.add_command(label="Trim", underline=0, accelerator = "alt-T", command=Self.Trim)
        editmenu.add_command(label="grow (percent)", underline=0, accelerator = "alt-G", command=Self.Grow)
        editmenu.add_command(label="Grow (fixed)", underline=0, accelerator = "alt-shift-G", command=Self.Grow)
        editmenu.add_command(label="Toggle Grid", underline=7, accelerator = "ctrl-G", command=Self.Toggle_Grid)
        editmenu.add_command(label="Options", underline=0, accelerator = "ctrl-O", command=Self.Options)
        Self.menubar.add_cascade(label="Edit", menu=editmenu)
        
    def TopHelpMenu(Self) :
        helpmenu = Menu(Self.menubar, tearoff=0, relief=RAISED)
        helpmenu.add_command(label="About", command=Self.Menu.About)
        helpmenu.add_command(label="Help", command=Self.Menu.notdone)
        Self.menubar.add_cascade(label="Help", menu=helpmenu)

    def Options(Self, Event=None) :
        """ This routine displays and updates simulator parameters. """
        OptionsForm = Toplevel()
        OptionsForm.title("Options")

        SimDelay = IntVar()
        SimDelay.set(Self.SimDelay)
        SimDelayFrame = Frame(OptionsForm)
        SimDelayFrame.pack(side=TOP, fill=X)
        SimDelayLabel = Label(SimDelayFrame, text="simulation rate", width=15)
        SimDelayLabel.pack(side=LEFT)
        for Text, Delay in ((".1", 100), (".5", 500), ("1.0", 1000), ("2.0", 2000), ("3.0", 3000), ("5.0", 5000)) :
            SimDelayEntry = Radiobutton(SimDelayFrame, text=Text, variable=SimDelay, value=Delay)
            SimDelayEntry.pack(side=LEFT)

        SwitchMarker = IntVar()
        SwitchMarker.set(Self.SwitchMarker)
        SwitchMarkerFrame = Frame(OptionsForm)
        SwitchMarkerFrame.pack(side=TOP, fill=X)
        SwitchMarkerLabel = Label(SwitchMarkerFrame, text="switch markers", width=15)
        SwitchMarkerLabel.pack(side=LEFT)
        for Text, Value in (("off", 0), ("on", 1)) :
            SwitchMarkerEntry = Radiobutton(SwitchMarkerFrame, text=Text, variable=SwitchMarker, value=Value)
            SwitchMarkerEntry.pack(side=LEFT)

        Scale = IntVar()
        Scale.set(Self.Scale)
        ScaleFrame = Frame(OptionsForm)
        ScaleFrame.pack(side=TOP, fill=X)
        ScaleLabel = Label(ScaleFrame, text="scale factor", width=15)
        ScaleLabel.pack(side=LEFT)
        for Size in (1, 2, 3, 4) :
            ScaleEntry = Radiobutton(ScaleFrame, text=str(Size), variable=Scale, value=Size)
            ScaleEntry.pack(side=LEFT)

        LineWidth = IntVar()
        LineWidth.set(Self.LineWidth)
        LineWidthFrame = Frame(OptionsForm)
        LineWidthFrame.pack(side=TOP, fill=X)
        LineWidthLabel = Label(LineWidthFrame, text="line width", width=15)
        LineWidthLabel.pack(side=LEFT)
        for Size in (1, 2, 3, 4) :
            LineWidthEntry = Radiobutton(LineWidthFrame, text=str(Size), variable=LineWidth, value=Size)
            LineWidthEntry.pack(side=LEFT)

        FontSize = StringVar()
        FontSize.set(Self.FontSize)
        FontSizeFrame = Frame(OptionsForm)
        FontSizeFrame.pack(side=TOP, fill=X)
        FontSizeLabel = Label(FontSizeFrame, text="font size", width=15)
        FontSizeLabel.pack(side=LEFT)
        for Size in ("auto", "12", "14", "16", "18", "24", "36") :
            FontSizeEntry = Radiobutton(FontSizeFrame, text=Size, variable=FontSize, value=Size)
            FontSizeEntry.pack(side=LEFT)

        ButtonMode = StringVar()
        ButtonMode.set(Self.ButtonMode)
        ButtonModeFrame = Frame(OptionsForm)
        ButtonModeFrame.pack(side=TOP, fill=X)
        ButtonModeLabel = Label(ButtonModeFrame, text="button mode", width=15)
        ButtonModeLabel.pack(side=LEFT)
        for Mode in ("icon", "name", "icon+name") :
            ButtonModeEntry = Radiobutton(ButtonModeFrame, text=Mode, variable=ButtonMode, value=Mode)
            ButtonModeEntry.pack(side=LEFT)

        ColorMode = StringVar()
        ColorMode.set(Self.ColorMode)
        ColorModeFrame = Frame(OptionsForm)
        ColorModeFrame.pack(side=TOP, fill=X)
        ColorModeLabel = Label(ColorModeFrame, text="EPS output", width=15)
        ColorModeLabel.pack(side=LEFT)
        for Mode in ("color", "mono") :
            ColorModeEntry = Radiobutton(ColorModeFrame, text=Mode, variable=ColorMode, value=Mode)
            ColorModeEntry.pack(side=LEFT)

        OptionsVars = (SimDelay, SwitchMarker, Scale, LineWidth, FontSize, ButtonMode, ColorMode)
        ButtonFrame = Frame(OptionsForm)
        ButtonFrame.pack(side=TOP, fill=X)
        OKButton = Button(ButtonFrame, text="OK", command=(lambda Vars=OptionsVars,
                                                           Win=OptionsForm:
                                                           Self.Update_Options(Vars, Win)))
        OKButton.pack(side=LEFT, expand=YES, fill=X)
        CancelButton = Button(ButtonFrame, text="Cancel", command=OptionsForm.destroy)
        CancelButton.pack(side=RIGHT, expand=YES, fill=X)

    def Update_Options(Self, Vars, Window) :
        """ This routine collects the results of the options command and updates
        all the the information. """
        if Vars[0].get() <> Self.SimDelay :
            Self.SimDelay = Vars[0].get()
        if Vars[1].get() <> Self.SwitchMarker :
            Self.SwitchMarker = Vars[1].get()
        if int(Vars[2].get()) <> Self.Scale :
            Self.Scale = int(Vars[2].get())
            Self.Canvas.Update_Fonts(Self.FontSize, Self.Scale)
            Self.Canvas.Resize()
        if Vars[3].get() <> Self.LineWidth :
            Self.LineWidth = Vars[3].get()
            Self.Canvas.Refresh()
        if Vars[4].get() <> Self.FontSize :
            Self.FontSize = Vars[4].get()
            Self.Canvas.Update_Fonts(Self.FontSize, Self.Scale)
            Self.Canvas.Refresh()
        if Vars[5].get() <> Self.ButtonMode :
            Self.ButtonMode = Vars[5].get()
            Self.Menu.Destroy_Menu()
            Self.Menu.Build_Menu(Self)
        if Vars[6].get() <> Self.ColorMode :
            Self.ColorMode = Vars[6].get()
        Window.destroy()
        Self.Store_Init_File()

    def Edit_Comment(Self, Event=None) :
        """ This routine displays and updates the file comment. """
        CommentForm = Toplevel()
        CommentForm.title("File Description")

        Comment = StringVar()
        if Self.Comment :
            Comment.set(Self.Comment)
        CommentFrame = Frame(CommentForm)
        CommentLabel = Label(CommentFrame, text="file description", width=20)
        CommentEntry = Entry(CommentFrame, width=50)
        CommentEntry.config(textvariable=Comment)
        CommentFrame.pack(side=TOP, fill=X)
        CommentLabel.pack(side=LEFT)
        CommentEntry.pack(side=RIGHT, expand=YES, fill=X)

        ButtonFrame = Frame(CommentForm)
        ButtonFrame.pack(side=TOP, fill=X)
        OKButton = Button(ButtonFrame, text="OK", command=(lambda Var=Comment,
                                                           Win=CommentForm :
                                                           Self.Update_Comment(Var, Win)))
        OKButton.pack(side=LEFT, expand=YES, fill=X)
        CancelButton = Button(ButtonFrame, text="Cancel", command=CommentForm.destroy)
        CancelButton.pack(side=RIGHT, expand=YES, fill=X)

    def Update_Comment(Self, Var, Win) :
        """ This routine updates the comment. """
        if Self.Comment <> Var.get() :
            Self.Comment = Var.get()
            Self.Update_Title(Modified = True)
        Win.destroy()

    def Load_Init_File(Self) :
        """ This routine loads the current options settings to the options.ini file
        (if it exists). """
        Path = os.getcwd()     
        if "options.ini" in os.listdir(Path) :
            File = open(Path + "/options.ini", 'r')
            for Line in File.readlines() :
                Line = Line.strip()
                if len(Line) > 0 and ':' in Line :
                    I = Line.find(':')
                    Keyword = Line[:I]
                    Value = Line[I+2:]
                    exec("Self.%s = %s" % (Keyword, Value))
            File.close()

    def Store_Init_File(Self) :
        """ This routine saves the current options settings to the options.ini file
        in the binary directory. """
        FileName = os.getcwd() + "/options.ini"
        try :
            File = open(FileName, 'w')
        except :
            showerror("CompuCanvas", "Could not store options init file " + FileName)
        else :
            File.write("<CompuCanvas version %1.2f>\n\n" % (Self.Version))
            File.write("Width: %d\n" % (Self.Width))
            File.write("Height: %d\n" % (Self.Height))
            File.write("Scale: %d\n" % (Self.Scale))
            File.write("LineWidth: %d\n" % (Self.LineWidth))
            File.write("FontSize: \"%s\"\n" % (Self.FontSize))
            File.write("ButtonMode: \"%s\"\n" % (Self.ButtonMode))
            File.write("SimDelay: %d\n" % (Self.SimDelay))
            File.write("SwitchMarker: %d\n" % (Self.SwitchMarker))
            File.write("ColorMode: \"%s\"\n" % (Self.ColorMode))
            File.close()

    def New(Self, Event=None) :
        """ This routine begin a new design. """
        Self.Check_Modified()
        Self.Clear()

    def Clear(Self) :
        """ This routine clears the canvas to begin a new design. """
        Self.Comment = None
        Self.Objects = []
        Self.Selected = []
        for Key in Self.CurPaths.keys() :
            Self.CurPaths[Key] = None
        Self.Update_Title(FileName = None, Modified = False)
        Self.Canvas.Refresh()

    def Load(Self, Event=None, FilePath=None, ModCheck=True) :
        """ This routine allows a files to be selected and opened. """
        if ModCheck :
            Self.Check_Modified()
        if FilePath == None :
            if Self.CurPaths["sch"] :
                InitDir, InitFile = Self.CurPaths["sch"].rsplit('/', 1)
            else :
                InitDir, InitFile = ["../sch", '']
            FilePath = askopenfilename(parent=Self, initialdir=InitDir, initialfile=InitFile, filetypes=[("compucanvas schematic", ".sch")])
        if FilePath:
            Self.Clear()
            Self.CurPaths["sch"] = FilePath
            try :
                File = open(FilePath, 'r')
            except :
                showerror("Compucanvas", "Could not load file " + FilePath)
            else :
                for Line in File.readlines() :
                    Line = Line.strip()
                    if len(Line) > 0 :
                        if Line[0] == '<' :
                            if Line[1] == '"' :
                                Self.Comment = Line[2:-2]
                            elif Line.find('(') <> -1 :
                                Width, Height = Line[Line.find('(')+1:Line.find(')')].split(',')
                                Self.Width = int(Width)
                                Self.Height = int(Height)
                                Self.Canvas.Resize()
                        elif '(' in Line :
                            I = Line.find('(')
                            Line = Line[:I+1] + "Self.Canvas, " + Line[I+1:]
                            NewObject = eval(Line)
                            Self.Objects.append(NewObject)
                Self.Update_Circuit()
                Self.Canvas.Refresh()
                FileName = FilePath.rsplit('/', 1)[1]
                Self.Update_Title(FileName = FileName, Modified = False)

    def Save(Self, Event=None, FilePath=None) :
        """ This routine saves the canvas contents to the specified file path or
        the current file path. It also updates the current file path and title.
        If no file path is specified, it reverts to a "Save as". """
        if FilePath == None and Self.CurPaths["sch"] == None :
            Self.SaveAs(Event)
        else :
            if FilePath == None :
                FilePath = Self.CurPaths["sch"]
            try :
                File = open(FilePath, 'w')
            except :
                showerror("CompuCanvas", "Could not save file " + FilePath)
            else :
                Self.CurPaths["sch"] = FilePath
                File.write("<CompuCanvas version %1.2f (%d,%d)>\n" % (Self.Version, Self.Width, Self.Height))
                if Self.Comment :
                    File.write('<"%s">\n' % Self.Comment)
                File.write("\n")
                for Object in Self.Objects :
                    Object.Write(File)
                File.close()
                FileName = FilePath.rsplit('/', 1)[1]
                Self.Update_Title(FileName = FileName, Modified = False)

    def SaveAs(Self, Event=None) :
        """ This routine saves the canvas contents to a user specified file. A dialog menu
        is opened. If a file path is selected, it is sent to "Save" to save the schemtic. """
        if Self.CurPaths["sch"] :
            InitDir, InitFile = Self.CurPaths["sch"].rsplit('/', 1)
        else :
            InitDir, InitFile = ["../sch", '']
        FilePath = asksaveasfilename(parent=Self, initialdir=InitDir, initialfile=InitFile, filetypes=[("compucanvas schematic", ".sch")])
        if FilePath:
            if '.' not in FilePath :
                FilePath += ".sch"
            Self.Save(FilePath=FilePath)

    def Write_EPS(Self, Event=None, FilePath=None) :
        """ This routine writes an encapsulated postscript file of the current canvas.
        If a FilePath is not specified in the parameters, a dialog menu is launched. """
        if "eps" not in os.listdir("../") :
            os.mkdir("../eps")
        if FilePath == None :
            if Self.CurPaths["eps"] :
                InitDir, InitFile = Self.CurPaths["eps"].rsplit('/', 1)
            elif Self.CurPaths["sch"] :
                InitDir = "../eps"
                InitFile = Self.CurPaths["sch"].rsplit('/', 1)[1]
                InitFile = InitFile[:-4] + ".eps"
            else :
                InitDir, InitFile = ["../eps", '']
            FilePath = asksaveasfilename(parent=Self, initialdir=InitDir, initialfile=InitFile, filetypes=[("postscript", ".eps")])
        if FilePath:
            if '.' not in FilePath :
                FilePath += ".eps"
            try :
                File = open(FilePath, 'w')
            except :
                showerror("CompuCanvas", "Could not write EPS file " + FilePath)
            else :
                Self.CurPaths["eps"] = FilePath
            Self.Canvas.Draw_Frame(False)
            Self.Canvas.postscript(file=FilePath, colormode=Self.ColorMode, rotate=False)
            Self.Canvas.Draw_Frame(True)

    def Write_Plot_EPS(Self, Event=None, FilePath=None) :
        """ This routine writes an encapsulated postscript file of the current timing diagram.
        If a FilePath is not specified in the parameters, a dialog menu is launched. """
        if "eps" not in os.listdir("../") :
            os.mkdir("../eps")
        if FilePath == None :
            if Self.CurPaths["plt"] :
                InitDir, InitFile = Self.CurPaths["plt"].rsplit('/', 1)
            else :
                InitDir, InitFile = ["../eps", '']
            FilePath = asksaveasfilename(parent=Self, initialdir=InitDir, initialfile=InitFile, filetypes=[("postscript", ".eps")])
        if FilePath :
            if '.' not in FilePath :
                FilePath += ".eps"
            try :
                File = open(FilePath, 'w')
            except :
                showerror("CompuCanvas", "Could not write plot EPS file " + FilePath)
            else :
                Self.CurPaths["eps"] = FilePath
            Self.TimePlot.postscript(file=FilePath, colormode=Self.ColorMode, rotate=False)

    def Create_PNG(Self, Event=None, FilePath=None, Scale="100%") :
        """ This routine converts an encapsulated postscript file
        into a png file. It replaces the font and scale spec and
        converts to a png format using a command line expression.
        If a FilePath is not specified in the parameters, a dialog
        menu is launched. Note: this currently works in Linux. """
        if FilePath == None :
            if Self.CurPaths["eps"] :
                InitDir, InitFile = Self.CurPaths["eps"].rsplit('/', 1)
            else :
                InitDir, InitFile = ["../eps", '']
            FilePath = askopenfilename(parent=Self, initialdir=InitDir, initialfile=InitFile, filetypes=[("postscript", ".eps")])
        if FilePath :
            TmpFilePath = FilePath[:-4] + "-Tmp.eps"
            PngFilePath = FilePath[:-4] + ".png"
            InFile = open(FilePath, 'r')
            OutFile = open(TmpFilePath, 'w')
            for Line in InFile.readlines() :
                if Line.find("NimbusSansL") <> -1 :
                    Line = Line.replace("NimbusSansL", "Helvetica")
                N = Line.find("scalefont")
                if N <> -1 :
                    OldSize = Line[N-4:N].strip()
                    NewSize = str(int(OldSize) / 3 * 4)
                    Line = Line.replace(OldSize, NewSize)
                OutFile.write(Line)
            InFile.close()
            OutFile.close()
            CmdStr = "convert -alpha off +antialias -scale %s -density 128 %s %s" % (Scale, TmpFilePath, PngFilePath) 
            P = subprocess.Popen(CmdStr, shell=True)
            P.wait()
            os.remove(TmpFilePath)

    def Dir_Create_PNG(Self, Event=None) :
        """ This routine accepts a directory locatation contaning sch files.
        Each file is loaded, trimed,writen as an EPS file, and converted to a
        PNG file. Then the EPS file is removed. Thw files are converted at 50%
        scale. """
        Dir = askdirectory(parent=Self, initialdir="../sch", title='Select a directory')
        if Dir <> "" :
            for FileName in os.listdir(Dir) :
                if len(FileName) > 4 and FileName[-4:] == ".sch" :
                    SchPath = Dir + '/' + FileName
                    print SchPath
                    EpsPath = SchPath[:-4] + ".eps"
                    Self.Load(FilePath=SchPath, ModCheck=False)         # load schemtic
                    Self.Trim()                                         # trim schematic
                    Self.Tick()                                         # tick the clock
                    tkMessageBox.showinfo("Click OK to continue\n")     # let window system update
                    Self.Write_EPS(FilePath=EpsPath)                    # create EPS
                    Self.Create_PNG(FilePath=EpsPath, Scale="50%")      # create PNG
                    os.remove(EpsPath)                                  # remove EPS file

    def Trim(Self, Event=None) :
        """ This routine trims the canvas to exactly fit the schematic.
        It also places the upper left corner at (0,0). This routine also
        resizes the canvas. The bbox function appears to pad the objects
        (often by 4 pixels). So its results are dialated to a multiple
        of ten. """
        if Self.Objects :
            Self.Canvas.Draw_Frame(False)   # undraw page frame
            Grid = Self.Grid
            if Grid :
                Self.Toggle_Grid()
            Xmin, Ymin, Xmax, Ymax = Self.Canvas.bbox("all")
            Xmin /= Self.Scale
            Ymin /= Self.Scale
            Xmax /= Self.Scale
            Ymax /= Self.Scale
            Xmin -= Xmin % 10
            Ymin -= Ymin % 10
            Xmax += 10 - (Xmax % 10)
            Ymax += 10 - (Ymax % 10)
            Self.Width = Xmax - Xmin
            Self.Height = Ymax - Ymin
            for Object in Self.Objects :
                Object.Base[0] -= Xmin
                Object.Base[1] -= Ymin
            Self.Update_Title(Modified = True)
            Self.Canvas.Resize()
            if Grid :
                Self.Toggle_Grid()

    def Grow(Self, Event=None) :
        """ This routine grows the canvas. It "squares up" the canvas
        by increasing the width by 20% of the height and increasing the
        height by 20% of the width. The current objects are then centered
        with the resized canvas. If the key select is capitalizes, it adds
        a minimum step to all borders. """
        if Event and Event.keysym.isupper() :
            Xadjust = Yadjust = 20
        else :
            Xadjust = Self.Height / 5   # X is 20% of Y
            Yadjust = Self.Width / 5    # Y is 20% of X
        Self.Width += Xadjust
        Self.Height += Yadjust
        if Self.Objects :
            Xadjust /= 2
            Yadjust /= 2
            if Xadjust % 10 < 5 :
                Xadjust -= Xadjust % 10
            else :
                Xadjust += 10 - (Xadjust % 10)
            if Yadjust % 10 < 5 :
                Yadjust -= Yadjust % 10
            else :
                Yadjust += 10 - (Yadjust % 10)   
            for Object in Self.Objects :
                Object.Base[0] += Xadjust
                Object.Base[1] += Yadjust
            Self.Update_Title(Modified = True)
        Self.Canvas.Resize()

    def Update_Title(Self, FileName="", Modified=None) :
        """ This routine updates the window title bar with the filename (if known)
        and the modification star (if modified). """
        if FileName <> "" :
            Self.FileName = FileName
        if Modified <> None :
            Self.Modified = Modified
        String = "CompuCanvas %.2f" % Self.Version
        if Self.FileName :
            String = Self.FileName + "     " + String
        if Self.Modified :
            String = '*' + String
        Self.title(String)

    def Check_Modified(Self) :
        """ This routine checks if the current design is modified.
        If so, it asks if it should be saved. """
        if Self.Modified :
            Save = askyesno(title="Warning", message="Current design is not saved. Do you want to save it?")
            if Save :
                Self.Save()

    def Mark_Mouse(Self, Event=None) :
        """ This routine sets the mouse position. """
        (Self.MouseX, Self.MouseY) = Self.Canvas.Adjust_Position(Event.x, Event.y)

    def Update_DragBox(Self, Event=None) :
        """ This routine draws the dragbox for region select. """
        if Self.MouseX and Self.MouseY :
            (EventX, EventY) = Self.Canvas.Adjust_Position(Event.x, Event.y)
            if Self.MouseX > EventX :
                X1, X2 = EventX, Self.MouseX
            else :
                X1, X2 = Self.MouseX, EventX
            if Self.MouseY > EventY :
                Y1, Y2 = EventY, Self.MouseY
            else :
                Y1, Y2 = Self.MouseY, EventY
            Self.Canvas.Clear_DragBox()
            Self.Canvas.Draw_DragBox(X1, Y1, X2, Y2)

    def Select(Self, Event=None) :
        """ This routine selects a clicked object (or objects in a region). First all
        currently selected objects are dehighlighted. If the clicked object is already
        selected, it is highlighted. It replaces all currently selected objects. It also
        clears the mouse position. """
        if Self.MouseX and Self.MouseY :
            Self.Highlight_Selected(False)
            Self.Selected = []
            (EventX, EventY) = Self.Canvas.Adjust_Position(Event.x, Event.y)
            if EventX == Self.MouseX and EventY == Self.MouseY :
                Objects = Self.Canvas.Find_Clicked_Object(EventX, EventY)
            else :
                Objects = Self.Canvas.Find_Region_Objects(Self.MouseX, Self.MouseY, EventX, EventY)
            for Object in Objects :
                Object.Highlight(True)
                Self.Selected.append(Object)
            Self.Canvas.Clear_DragBox()
            Self.MouseX = None
            Self.MouseY = None

    def Add_Select(Self, Event=None) :
        """ This routine adds a clicked object (or objects in a region) to selected.
        If it is already selected, it is unhighlighted and removed from the selection.
        It also clears the mouse position. """
        if Self.MouseX and Self.MouseY :
            (EventX, EventY) = Self.Canvas.Adjust_Position(Event.x, Event.y)
            if EventX == Self.MouseX and EventY == Self.MouseY :
                Objects = Self.Canvas.Find_Clicked_Object(EventX, EventY)
            else :
                Objects = Self.Canvas.Find_Region_Objects(Self.MouseX, Self.MouseY, EventX, EventY)
            for Object in Objects :
                if Object not in Self.Selected :
                    Self.Selected.append(Object)
                    Object.Highlight(True)
                else :
                    Self.Selected.remove(Object)
                    Object.Highlight(False)
            Self.Canvas.Clear_DragBox()
            Self.MouseX = None
            Self.MouseY = None

    def Select_All(Self, Event=None) :
        """ This routine selects all objects. First all selected objects are unhighlighted.
        then all objects are included in the selected list and highlighted. """
        Self.Highlight_Selected(False)
        Self.Selected = copy(Self.Objects)
        Self.Highlight_Selected(True)

    def Highlight_Selected(Self, On) :
        """ This routine highlights or unhighlights all objects in the current selected list. """
        for Object in Self.Selected :
            Object.Highlight(On)

    def Cut(Self, Event=None) :
        """ This routine removes selected objects from the object list and adds them
        to the (local) clipboard. """
        Self.Highlight_Selected(False)
        for Object in Self.Selected :
            Self.Objects.remove(Object)
        Self.Clipboard = Self.Selected
        Self.Selected = []
        Self.Update_Circuit()

    def Copy(Self, Event=None, Offset=True) :
        """ This routine copies selected objects and adds them to the (local) clipboard.
        If Offset is true, an offset is added to base of the copied objects. """
        Self.Highlight_Selected(False)
        Self.Clipboard = []
        for Object in Self.Selected :
            NewObj = eval(Object.Clone())
            if Offset :
                for I in (0,1) :
                    NewObj.Base[I] += 10
            Self.Clipboard.append(NewObj)
        Self.Highlight_Selected(True)

    def Paste(Self, Event=None) :
        """ This routine adds objects on the clipboard to the object list. It also selects
        and highlights them. It also replaces the clipboard objects with new copies. """
        Self.Highlight_Selected(False)
        Self.Selected = copy(Self.Clipboard)
        Self.Highlight_Selected(True)
        Self.Objects += Self.Clipboard
        Self.Copy(Offset=False)
        Self.Update_Circuit()

    def Move_Up(Self, Event=None) :
        """ This routine moves selected objects right by one grid. """
        for Object in Self.Selected :
            Object.Move((0,-1))
        Self.Update_Circuit()

    def Move_Down(Self, Event=None) :
        """ This routine moves selected objects right by one grid. """
        for Object in Self.Selected :
            Object.Move((0, 1))
        Self.Update_Circuit()

    def Move_Left(Self, Event=None) :
        """ This routine moves selected objects right by one grid. """
        for Object in Self.Selected :
            Object.Move((-1,0))
        Self.Update_Circuit()

    def Move_Right(Self, Event=None) :
        """ This routine moves selected objects right by one grid. """
        for Object in Self.Selected :
            Object.Move((1,0))
        Self.Update_Circuit()

    def Delete(Self, Event=None) :
        """ This routine deletes the selected objects. """
        for Object in Self.Selected :
            Self.Canvas.delete(Object.ID)
            Self.Objects.remove(Object)
        Self.Selected = []
        Self.Update_Circuit()

    def Properties(Self, Event=None) :
        """ This routine edits the properties of each selected object. First
        all selected objects are unhighlighted. Then they are processed one
        at a time using an object calling chain. """
        if Self.Selected :
            Self.PropMod = True
            Self.Highlight_Selected(False)
            Self.Selected[0].Properties(Self.Selected[1:])

    def Properties_Complete(Self) :
        """ This routine completes the property editing mode. The circuit is
        reextracted and selected objects are highlighted. PropMod is cleared. """
        Self.Update_Circuit()
        Self.Canvas.Refresh()
        Self.PropMod = False

    def Toggle_Source(Self, Event=None) :
        """ This routine toggle the source on a clicked object (if possible).
        Currently only labels have sources that can toggle. """
        (EventX, EventY) = Self.Canvas.Adjust_Position(Event.x, Event.y)
        Objects = Self.Canvas.Find_Clicked_Object(EventX, EventY)
        for Object in Objects :
            if Object.Type == "LABEL" :
                Object.Toggle_Source()

    def Toggle_Grid(Self, Event=None) :
        """ This routine draws (or undraws) a grid on the canvas view. """
        if Self.Grid :
            Self.Canvas.Erase_Grid()
        else :
            Self.Canvas.Draw_Grid()

    def Place(Self, Event=None) :
        """ This routine places the active prototype object type on the canvas. """
        (EventX, EventY) = Self.Canvas.Adjust_Position(Event.x, Event.y)
        Base = Self.Canvas.Find_Closest_Grid(EventX, EventY)
        NewObject = Self.Mode.ObjectType(Self.Canvas, Base)
        Self.Mode.Exit()
        Self.Mode = NormalMode(Self)
        Self.Objects.append(NewObject)
        Self.Selected = [NewObject]
        Self.Update_Circuit()

    def Connect(Self, Event=None) :
        """ This routine connects two terminals through multiple segments. If no
        wire exists in the connect mode, the point is checked in the contact
        list and a new wire is created. Otherwise connect mode is exited. Once
        a wire is created, additional click points form either (a) a new segment
        when a non-contact grid point is selected, or (b) the second terminal
        finishing the wire object. """
        (EventX, EventY) = Self.Canvas.Adjust_Position(Event.x, Event.y)
        Point = Self.Canvas.Find_Closest_Grid(EventX, EventY)
        if Self.Mode.NewWire == None :
            if tuple(Point) in Self.Sim.Contacts :
                Self.Mode.NewWire = Self.Mode.ObjectType(Self.Canvas)
                Self.Mode.NewWire.Points = [Point]
            else :
                Self.Mode.Exit()
                Self.Mode = NormalMode(Self)
        else :
            if Event.state & 4 == 0 : # HV lock (unlocked with <ctrl>)
                Last = Self.Mode.NewWire.Points[-1]
                if abs(Point[0] - Last[0]) > abs(Point[1] - Last[1]) :
                    Point[1] = Last[1]
                else :
                    Point[0] = Last[0]
            if tuple(Point) in Self.Sim.Contacts :
                Self.Mode.NewWire.Finish(Point)
                Self.Objects.append(Self.Mode.NewWire)
                Self.Mode.Exit()
                Self.Mode = NormalMode(Self)
                Self.Selected = [Self.Objects[-1]]
                Self.Update_Circuit()
            else :
                Self.Mode.NewWire.Add_Point(Point)

    def Remove_Point(Self, Event=None) :
        """ This routine removes a point from the wire being drawn. """
        if Self.Mode.NewWire :
            Self.Mode.NewWire.Remove_Point()

    def Update_Circuit(Self) :
        """ This routine creates a new terminal list from all objects. """
        Self.Update_Title(Modified = True)
        Self.Dot_Wire_Adjust()
        Self.Sim.Derive_Nodes(Self.Objects)
        Self.Sim.Find_Node_State()
        Self.Canvas.Refresh()

    def Dot_Wire_Adjust(Self) :
        """ This routine looks for intersections of dot objects and wire
        objects. If an intersection is found at wire endpoints, no action is
        taken. If the dot falls on the wire's drawn segments, the old wire
        is replaced by two new wires with a shared endpoint at the dot. """
        for Dot in Self.Objects :
            if Dot.Type == "DOT" :
                for Wire in copy(Self.Objects) :
                    if Wire.Type == "WIRE" and Wire.Within(Dot.Base) :
                        NonContact = True
                        for N in (0,1) :
                            if Dot.Base[0] == Wire.Terminals[N].Contact[0] + Wire.Base[0] and Dot.Base[1] == Wire.Terminals[N].Contact[1] + Wire.Base[1] :
                                NonContact = False
                        if NonContact and Wire.Intersect(Dot.Base) :
                            Self.Objects += Wire.Split(copy(Dot.Base))
                            if Wire in Self.Selected :
                                Self.Selected.remove(Wire)
                                Self.Selected += Self.Objects[-2:]
                                Self.Selected[-1].Highlight(True)
                                Self.Selected[-2].Highlight(True)
                            Self.Objects.remove(Wire)
                            Self.Canvas.delete(Wire.ID)

    def Next_Label(Self) :
        """ This routine finds the first unused letter in
        current labels. """
        Labels = []
        for Object in Self.Objects :
            if Object.Type == "LABEL" :
                Labels.append(Object.Text)
        for Letter in "ABCDEFGHIJKLMNOPQRSTUVWXYZ" :
            if Letter not in Labels and '-' + Letter not in Labels :
                return Letter
        return "NewLabel"

    def RunStop(Self, Event=None) :
        """ This routine toggles the simulation run mode. The run variable is
        toggled and run/stop indicator is adjusted. """
        if Self.Run :
            Self.Run = False
        else :
            Self.Run = True

    def SimLoop(Self) :
        """ This routine is the primary simulation event sequence. It checks the run
        variable and conditionally executes a simulation cycle. Then it schedules the
        next cycle after the appropriate delay. """
        Status = Self.Run and Self.Objects and Self.Mode.__class__ == NormalMode and not Self.PropMod
        Self.Menu.RunStatus.Update(Status)
        if Status :
            Self.Tick()
        Self.after(Self.SimDelay, Self.SimLoop)

    def Tick(Self, Event=None) :
        """ This routine executes a single tick of the clock. """
        Self.Cycle += 1                                                 # advance clock
        for Object in Self.Objects :
            if Object.Type == "CLOCK" :
                Object.Set_Source(Self.Cycle, Object.Depth, Object.Delay)   # set terminal value given Cycle num)
        Self.Sim.Simulate_Objects(Self.Objects)
        Self.Sim.Find_Node_State()
        if Self.PlotWin :
            Self.Plot_Tick()
        Self.Canvas.Refresh()

    def Plot(Self, Event=None) :
        """ This routine prepares to plot labels and clocks
        in a timing diagram. """
        PlotForm = Toplevel()
        PlotForm.title("Plot")
        SelectFrame = Frame(PlotForm)
        SelectFrame.pack(side=TOP, fill=X)
        Self.Sbar = Scrollbar(SelectFrame, orient=VERTICAL)
        Self.ListBox = Listbox(SelectFrame, yscrollcommand=Self.Sbar.set, exportselection=0,
                               selectmode=SINGLE, font=("Helvetica", 12, "bold"))
        Self.Sbar.config(command=Self.ListBox.yview)
        Self.Sbar.pack(side=RIGHT, fill=Y)
        Self.ListBox.pack(side=LEFT, fill=BOTH, expand=1)
        Labels = []
        for Object in Self.Objects :
            if Object.Type == "LABEL" and Object.Text not in Labels :
                Labels.append(Object.Text)
            elif Object.Type == "CLOCK" :
                Text = "CLK%d" % (Object.Phase)
                if Text not in Labels :
                    Labels.append(Text)
        Labels.sort()
        for Label in Labels :
            Self.ListBox.insert(END, Label)
        LabelButtonFrame = Frame(PlotForm)
        LabelButtonFrame.pack(side=TOP, fill=X)
        AddButton = Button(LabelButtonFrame, text="Add", command=Self.Add_Label)
        AddButton.pack(side=LEFT, fill=Y)
        RemoveButton = Button(LabelButtonFrame, text="Remove", command=Self.Remove_Label)
        RemoveButton.pack(side=RIGHT, fill=Y)
        Self.ListBox.bind("<Double-Button-1>", Self.Add_Label)

        PlotLength = IntVar()
        PlotLength.set(Self.PlotLength)
        PlotLengthFrame = Frame(PlotForm)
        PlotLengthFrame.pack(side=TOP, fill=X)
        Self.LabelList = []
        PlotLengthEntry = Scale(PlotLengthFrame, label="Plot length", from_=1, to=100, orient="horizontal", length=400)
        PlotLengthEntry.pack(expand=YES, fill=X)
        ButtonFrame = Frame(PlotForm)
        ButtonFrame.pack(side=TOP, fill=X)
        OKButton = Button(ButtonFrame, text="OK",  command=lambda: Self.Create_Plot_Pane(PlotLengthEntry, PlotForm))
        OKButton.pack(side=LEFT, expand=YES, fill=X)
        CancelButton = Button(ButtonFrame, text="Cancel", command=PlotForm.destroy)
        CancelButton.pack(side=RIGHT, expand=YES, fill=X)

    def Add_Label(Self, Event=None) :
        """ This routine selects a label and adds it to the ordered label list. """
        try:
            Selection = Self.ListBox.curselection()[0]
            N = int(Selection)
            Label = Self.ListBox.get(N, None)
            if Label in Self.LabelList :
                Self.LabelList.remove(Label)
            Self.LabelList.append(Label)
            Self.ListBox.delete(N, None)
            Self.ListBox.insert(0, Label)
            Self.ListBox.itemconfig(0, foreground="blue")
        except IndexError:
            pass

    def Remove_Label(Self, Event=None) :
        """ This routine selects a label and removes it from the ordered label list. """
        try:
            Selection = Self.ListBox.curselection()[0]
            N = int(Selection[0])
            Label = Self.ListBox.get(N, None)
            if Label in Self.LabelList :
                Self.LabelList.remove(Label)
            Self.ListBox.delete(N, None)
            Self.ListBox.insert(END, Label)
        except IndexError:
            pass

    def Create_Plot_Pane(Self, PlotLengthScale, PlotForm) :
        """ This routine creates the plotting pane. """
        Self.PlotLength = PlotLengthScale.get()
        PlotForm.destroy()
        Self.PlotWin = Toplevel()
        Self.PlotWin.title("Timing Diagram")
        PlotFrame = Frame(Self.PlotWin)
        PlotFrame.pack(side=TOP, fill=X)
        Width = Self.PlotLength * 10 + 50
        Height = len(Self.LabelList) * 50
        Self.TimePlot = Canvas(PlotFrame, bg="#FFFFFF", width=Width, height=Height)
        Self.TimePlot.pack(side=TOP, fill=X)
        Y = 50 * len(Self.LabelList) - 5
        Self.Labels = []
        Self.States = []
        Self.PlotStart = Self.Cycle
        for Label in Self.LabelList :
            for Object in Self.Objects :
                if Object.Type == "LABEL" and Object.Text == Label :
                    Self.Labels.append(Object)
                    Self.States.append(Object.Terminals[0].Node.State)
                if Object.Type == "CLOCK" and Label == "CLK%d" % (Object.Phase) :
                    Self.Labels.append(Object)
                    Self.States.append(Object.Terminals[0].Node.State)
            Self.TimePlot.create_line(40, Y-40, 50, Y-40, fill="red", width=2, tags="Labels")
            Self.TimePlot.create_text(25, Y-20, text=Label, font=("helvetica", 12, "bold"), anchor=CENTER, tags="Labels")
            Self.TimePlot.create_line(40, Y, 50, Y, fill="blue", width=2, tags="Labels")
            Y -= 50

    def Plot_Tick(Self) :
        """ This routine plots a tick result in the timing diagram. """
        Slot = (Self.Cycle - Self.PlotStart) % Self.PlotLength
        OldStates = Self.States
        Self.States = []
        Clock = []
        for Object in Self.Labels :
            if Object.Type == "LABEL" :
                Self.States.append(Object.Terminals[0].Node.State)
                Clock.append(0)
            if Object.Type == "CLOCK" :
                Self.States.append(Object.Terminals[0].Node.State)
                Clock.append(Object.Phase)
        Y = Ymax = 50 * len(Self.LabelList) - 5
        Tag = "S%d" % (Slot)
        try :
            Self.TimePlot.delete(Tag)
        except TclError:
            Self.PlotWin = None
            Self.TimePlot = None
            return 
        for N in xrange(len(Self.LabelList)) :
            X = 50 + Slot * 10
            if Clock[N] and OldStates[N] <> Self.States[N] :
                if Clock[N] == 1 :
                    Fill = "green"
                elif Clock[N] == 2 :
                    Fill = "purple"
                Self.TimePlot.create_line(X+10, 0, X+10, Ymax, width=2, fill=Fill, tags=Tag)
            if OldStates[N] == 0 :
                Self.TimePlot.create_line(X, Y, X+10, Y, width=2, tags=Tag)
                if Self.States[N] == 1 :
                    Self.TimePlot.create_line(X+10, Y, X+10, Y-40, width=2, tags=Tag)
            if OldStates[N] == 1 :
                Self.TimePlot.create_line(X, Y-40, X+10, Y-40, width=2, tags=Tag)
                if Self.States[N] == 0 :
                    Self.TimePlot.create_line(X+10, Y, X+10, Y-40, width=2, tags=Tag)
            if OldStates[N] in (2, 3, 4) :
                Fill = ("gray", "purple", "green")[OldStates[N] - 2]
                Self.TimePlot.create_rectangle(X, Y, X+10, Y-40, width=0, fill=Fill, tags=Tag)
            Self.TimePlot.create_line(X+10, 0, X+10, Ymax, width=4, fill="gray", tags="S%d" % (Slot + 1 % Self.PlotLength))
            Y -= 50

    def Update_State(Self) :
        """ This routine updates the node states and refreshes the canvas. """
        Self.Sim.Find_Node_State()
        Self.Canvas.Refresh()

    def Exit(Self, Event=None) :
        """ This routine exits CompuCanvas. """
        Self.Check_Modified()
        Self.quit()

######################### Menu Pane #########################

Load = """R0lGODlhEAAQAOcAAP///9iyLP///v788/777+LDYOPHa9iILf788tizLBl5
AP777q3NlKrLkUiRGkmRG1GXIVeYKFubKVygPF6fOl+hQ2KjRmOeMGOfMWWf
MWalSGilRmmiNGujN26kNG+kOG+lPXSrT3aoPHyvVYCzYYGzWoK1Z4W2ZoW2
aom4a466b5C8dJbBgJfCgpjCg57FiKHGiqHHiqXIjajLkKnKkEOPFTuLD7DO
ltiGLNiHLTiHCBt5ADN+BNmOM9mYNdmxMtm0NNqUNdqaNtqfNdqhNtqkNS2D
AdqqNtqrNdqrNtquNCuDAdqvNtqzNtu2POHCXuLDXyV/ACF9AOTHa+XKdObN
eenBc+rLbO3Kau7OiO7Pcu7Pdu/Qeu/SgO/The/UlPLVgfPalvTYXPTbnvTc
k/Tdp/XekfXelPXgn/bcbPbgfPbhrPbpyffhgffjhffji/jkjvjlkfjmlPjm
lfnonPnpofnppPrsrPrts/rw3Pvvufv26Pv36fv36vzyyPzz0Pz01/z22P36
8f368v767x99AB97AP788P788dqmNx57A/789v789/799P7+/f/89f/99v/9
+P/9+v/+/f/+/tqvNRV5AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACwAAAAAEAAQAAAI2wABCBxIsKDBgwWBPCnAEIqT
ABAhJkggsIkjAZQmCegzxYDHKlR+CKzECI4ZNGsINBqAiMACSUwEHhF0hgwY
LFe8dOGyRQsfJQKR7PmSxQobBAMQHHIpIInARAIiQXo0IJAcNyIyiAFURCCR
RX703KHzRk0aDyUi/BkiUMggPHbmgPjA4cKIBiEc+BAYhFCdOG063GDQYMaK
GBt0GAHQI0+ZMWEwMFCR4gQJEy4mRAGA40COAwck0JAB4wWKFhp47DAI4UEN
GxRYWCikAOHAJRUM1bYtUIoiSwMDAgAh/hBBZG9iZSBJbWFnZVJlYWR5ADs="""

Save = """R0lGODlhEAAQAOcAAPj7/ojAYjZrvMLcv/b5/WSMyDZqu3uj4jZru/H2/Hmi
4ff6/nqj43Sd3Ojw+jFgp/f6/TRntHqj4nqj4b3Q7Hih4Heg3tnl99vn9z5w
uzhruzZquuHs+Xuk4nuj4TdrujVqujhsuzlsvDptuztuvTVpuV6JyTVpuGWM
yGWNyWWNymaNyWeOyWiQzWmQymqTz2ySy22Sy22W03CZ1nGZ2HKb2XKc23Ob
2XOb2nSc2jVpt3Se3XWe3naf3Xaf3nef3jVptneg3zVptXii4HmYxzVot3qi
4Hqi4Xqi4jVotjVotTRms3qk4zRmsjRlsTNlsHuk43yayHybyX2m5n2n536o
536o6ICq6YGezISjzjNkr4qn0o6r1ZWw2pq13Z+64aS+5KnC563F6q3G67PK
7bbM7rjO77rQ7rvQ7zNkrr7S8MLV8jNjrcPV78XY8sfZ9Mnc9Mzd9dDf9tHf
9tHg9tHg99jl9jJjrNvm9zJiq9vn+N3o+N/p+N/p+TJiquLs+ePt+eXt+jJh
qunw+unx+urw+urx++vx++3y++7z++70++70/O/0/PD1/DJhqfL2/PP3/fP4
/TJhqPb6/fb6/jFhqDFgqPf7/vj6/jFgpv///wAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACwAAAAAEAAQAAAI/wAFCBSIAIEBAyJIjNCQAYgm
AXTqALhEoFEhRAQgHOJDwYQQAXOuUJpUIJGhRwAS/NFzhkIGAXKsJEhQgJAi
CAQc7MEzoY0OAXGqOHCwYlAkTIz6YLgwJEwSBHCocOAAaBEESIGW2rEAJgKC
N1MKpGABI4YLFAXS5viyxIAbKAyYdDjgQUEQHztsNPDixMAaBgwOHOigoIIF
Hg1w3OiixYAaCRIm0K3wo0eDGjRmcGGzAQ0SAKBDi5axJY8GMwoWBFjNejWB
F1kEhShzZMGA27hvE2iBRdIHMkZUt2ZNQIWUByDGiBHNHEAUIg82lDhRREmE
Jk/S3PHjyNKDTJUCAgAh/hBBZG9iZSBJbWFnZVJlYWR5ADs="""

RunStop = """R0lGODlhEAAQAOcAAP//////AP/QTv/NAADRAOEAAP/yAP+/AP/+AP7/AP7+
AOAAAP/zAP/0AP/1AP/YAP/gAP/5AADSAP/hAADQAOY1AOhDAOk2AOo6AOtY
APBdAPBfAPFqAPFsAPSPAPWnAPiSAPjeAPu3AOIXAP7JAP7iAP7oAP7tAP77
AP78AOECAOEBAP+7ANzBANfbANTTAP/STv/TAMfNAP/dAP/fAMPTAKvCAP/k
AP/lAP/oAP/pAP/qAP/rAKu/AJrCAA/PAAzYAAbTAP/6AP/9AATPAADTAP24
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACH5BAEAAAAALAAAAAAQABAAAAioAAEIHBAAgYKD
ChAgMCBAoEOCKEhw6GAkwsKGDgEMSCBCxYIFK0CkYJhRoxAMBVIWyKCDZMYB
ES6oLFBhh8uHATSoXPCBx82BAUpsGGHBw4MAPzUGCOAgxw0cBZMSLDhhAoQE
F0sSdNBDgoQfLYZIDSCDgFkCQCAwwIjTx1mzLxqwHZjAxtsiM+RqDUCDCAUK
QVycGJvgQYgaMUwEWFuSxYHHkCHDEBgQADs="""

Options = """R0lGODlhEAAQAOcAAP///5ubm6O96MfHx+Tk5Ji05lGD1ZaWlpOTk5KSkoeH
h4WFhYGk33Sc3nie3nmf3Xmf3nub0n+l34GBgXOc3oKCgoODg4Ol34SEhHGZ
3IWn4YaGhm+X24ep4YiIiIin2omJiYmp4YqKiout44yMjIyp2Iyu4o2NjY6O
jo+Pj2WR2ZKx5F2J0pSz5ZSz5pWVlVeH1Za05ZiYmE+B001/05mZmZm35pq2
5kt90Zu35pu455ycnJy35py5552dnZ25556enkl90aW+6aW/6afA6ajA6qvC
6qvD6qysrKzD6q3E6q7G67PI7LPI7bW1tbfM7bjM7rm5ubq6urrN7rrO7rvO
7ry8vLzP78DAwMLCwsPDw8TExMXFxcbGxkd70cjIyMzMzM/Pz9LS0tTU1NfX
193d3eLi4pi25ufn5+7u7kV50QAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACH5BAEAAAAALAAAAAAQABAAAAioAAEIHDgQyI4A
NQgqFOgjAIE0BxYqDGAGTRQEEgnKIBAmAQAUJERkPFCGS4oTSMgokPgCwZgv
VqRk2bJA4YgPCQaAGdBFCxYMFQiG6HGmhBgnIBQssDCBYAceV6iYiOBhg0QN
NqoQEcCEAQuJFwpMGRLjxxIIBhZKcAFFSIscShzMkPjgiYAVN5I0oJGRQxMd
BY5QCJIRgAoGRopk8FJYIAwDONQ0BhAQACH+EEFkb2JlIEltYWdlUmVhZHkA
Ow=="""

Trim = """R0lGODlhEAAQAMZeAEKLP0KMP0ONQESNQEaQQkeRQkmURUqVRUuWRkyXR06X
Sk2ZSE6ZSE+bSlCcSlGeS1KfTFqcVlueV1WiTlajT1akT1ekUF6hWmChXFim
UVmnUmKkX1qoUlupU1uqVFyqVF6qV1yrVF2rVV2sVWipY1+tVl+uV2CvWGGx
WWSwXGKxWWKyWWOzWmSzWm2xZ26xaGa2XHWwcW+zaWe3XWi5Xmm5X3ezc3mz
dGq7YGu8YHe4cWy9YXe5b362eWy+Ynu4dHm5cm2/Ym2/Y3q6c3i7cHm7cW7A
Y3DDZXPCaXHDZYa9gHPGZ4W/fXTHaIm/g4rBhIvBho3Dh4fKfpLGjJPHjJTH
jYzMg4rNgIvOgZDPhpjNjqDQmJ/Rl6fWnv//////////////////////////
////////////////////////////////////////////////////////////
/////////////////////////////////////////////////yH+EUNyZWF0
ZWQgd2l0aCBHSU1QACH5BAEKAH8ALAAAAAAQABAAAAeVgH9NS3+FhocrJ4JY
WUl/RocwfypFPB9LV11WSD6HLSpEVUMcf0dSWlw5hyhUTEAahkJcWzWHJlNR
IIZBOzg0M4ciHRoWE4fHyMnKy8zNyCwoJiMehw8ODAkHhilPULCGDT032oUm
Oj9KFocLNiQXBX8lOk4vxYcIChsxEgMhMi4V/kA4ZOAPAQwRAvzJQIGZAAB/
AgEAOw=="""

Grow = """R0lGODlhEAAQAMZfAEGKPkKLP0KMP0ONQESNQEWOQUWPQUaQQkeRQ0mURUyU
R0qVRU2ZSE6ZSE+bSlCcSlKfTFShTlWiTlajT16gWlekUFelUFimUVumVFmn
UmOkX1qoUlqoU2GlXVupU2SmXlyqVGSnX1yrVF2sVWeoYl6tVl+tVl+uV2Cv
WGGwWGGxWWKxWWKyWXCtbGOzWnGtbWS0W3Oubme0X2a2XHG0anO0bGi5Xmm5
X3O1bHS1bnmzdXO2bHO2bWq7YGu8YHW4bnq3c22/Y27AY329dX69dnDDZYO7
fXHDZXLEZnLFZnXFa3PGZ3THaIbAfom/g4rAg4rAhILEeYTHe5DFionJgIfK
fpPHjJTIjYzMg5XJjpvRkqTTm6XWnafWnqnYof//////////////////////
////////////////////////////////////////////////////////////
/////////////////////////////////////////////////yH+EUNyZWF0
ZWQgd2l0aCBHSU1QACH5BAEKAH8ALAAAAAAQABAAAAeOgExMS0lHf4eIMzAs
KCVMXl2GiJMyV1kgS1xaVUKThytETVMbSEpYW1I+kypDVjwZFkV/QVRRN5Mn
PzgZfxKIPTaefx68wsbHyMnKy8InI8IPDYgufyc7NMWHDiEkC38IKdZQORWT
DB06GgoGJk9ANeWeCRQfLQQiTkYYxwcvMQIcLkyIAEHYgQIDAgAIBAA7"""

Exit = """R0lGODlhEAAQAOcAAP////90d/heYfhfYvddYPtpbOIdH/dbXf93evZYW/VW
WfNRVOovMfBHSu5BQ+5AQus5PO0/Qew7Puw8Pu09P+0+QOw6Pe49P+s4O+s3
Oe5BRO5CRe9DRu9ER+9FSPBGSPBGSes0NvE/QfJFR/JMT/JNT+kyNfRISvRN
UPVOUPVWWOktL/VXWvZKTPZVWPZWWfZYWugwMvZZW/ZZXPZaXPZaXectMOYq
LPdcXvdcX+UnKvdfYeUlJ+QiJPhgY/hhZPhiZPlhY/ljZvloavphY/piZfpj
ZvpkZ/pqbfpsbvtnaeMgIvttb/ttcPxrbvxzdv1ucf1vcv1xdP10d/15fP5x
dP52ef56ff5+gf9UV/9WWf9YW/9bXv9iY/9naP9sbv9ucP9wcf9wcv9wc/9x
dP9ydP9zdv90duMfIf92ef93efdVV/95e/95fP95ff96ff97fv98fv9+gP9+
gf+Chv+HiuIcHgAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACH5BAEAAAAALAAAAAAQABAAAAjBAAEIHEhQoI+C
AAIEQDjggAKCARBYiUJQwI8hC0oIjDgHi5QCAgkMYBJnQIMNAMxUuVKHSoEg
OQQkebMDhIaBUJxMofMECAEkbAh4cFBQiZAmcuC0UYOjwwOEAATUcMNlyxkO
EaACmHEgjZYsYj5AgJqAxpEyY8B8WXMhRkEWMoyQgeFgAgovKULoGKggQZEw
Lx5IyGBiRJcTK5YIXKCCiIsIFgTauCGiBQMDAxuQqDB2II8elwtSwIAQjR2t
qAUGBAAh/hBBZG9iZSBJbWFnZVJlYWR5ADs="""

Car = """R0lGODlhEAAQAOcAAP///ykpKUyJxkqHxBxLk2RkZEiEwkWAvygxPZ6/5qC/
5T4+PikqLCksMCkyPio7UipYnC03QjBlqTFEWTRqrTVObjhvsDt0tDw8PCkq
Kz5ETD94uEBAQEJKVUJ8vENITigsMSU0TElPVkligCRTmSNUmk1kgVBnglCL
xlqP01xcXFyQ0yBQl2CU1GJyhx5NlWR1h2SMwmWBo2aY1muGpmxsbGyd2G2I
pm6e2XB9i3CJp3Kh2nSi2nek23mm23qRr3uMoXun3H2Yu32o3H+q3YGr14Ks
3oWv34av34eu2Yiw34ix2oyu2Iyz4Y604Y+gtZCfr5C24pa34pq85Zy85Z29
5RxAc19yhqTC56jE6K7J57PN6rTN6rfP6rjR67rR7LvS7bzT7cHW7ho4YAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACH5BAEAAAAALAAAAAAQABAAAAiiAAEIHEiwoMGD
CBMqXChAwAADBzxsuGCBgkEBS7x86bKFixYmEgoKQAHlSgIdNE4ggTGGoMMi
OW5EkWFiRA8XFayUYAFgQBIxCZwoMRKEB44hCqjEeAHAQJgpTY4Q8bHDxowW
K1JIIUHgAJgnH0QAyYJFQZUfGjoIgUBgQgQGCzg0eOAAAYIMCzCACEFAYIAC
NQIQ/KtC8MAAiA37Tbyw8cKAACH+EEFkb2JlIEltYWdlUmVhZHkAOw=="""

Stop = """R0lGODlhEAAQAOcAAP////+Jaf+DZf9+X/95Wv9pS/+ZeP+dfP9zVf9wT/9u
T/+Tc/+Obv9kRv9eQf9eQP9VN91ZUf+igd5iWt5lXt9pY+BqY+BsZeBxbeJ+
eeN4ceRsY+R1bOWEfeWIg+aKheaLhueBeeeRi+iLg+iWk+iXkuiYlOmXkumY
kuqTjOqaleqdmeqemOugm+uineujnuyloO2fmO6tpu+spfCxqfKLffSUhvaW
ifeNfflgSPpCJfqOffqThPqXifqllvutnvutn95eVPvDtPyQf/yyovzIt/zN
vPzRwfzTwvzUxP1GKP1lSv5JLP9LLP9PMP9PMf9QMv9QM9xbU/9ZO/9ZPP9a
PP9cP9tVTdpWTv9jRtpPR/9nSP9nTtlLRNlLQ/9uUP9vUNhIQNhHPv90Vf90
Wv94WddIQddIQP9+YP9/X/+AYP+BZ9ZBOdU+Nf+NbdQ0LP+Ob/+Pc/+Qb9Iu
Jf+TdP+TeP+VhP+Ydv+YeNIrI/+Zh/+bitEuJf+efP+gff+gfv+hhf+hkPu6
rP+lgf+nhP+nhf+qhv+qmP+rif+ukf+vi/+7nP+7pv++q/+/oP/Co//Esf/N
uP/Ouf/Pu//Rvf/WwdEqIgAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACH5BAEAAAAALAAAAAAQABAAAAjaAAEIHEiwYMEV
L2C4aMHiBAgPBk3QQFKJ0iRJRWJkKEhCRpJHihAR+pNISAoMA0vMOOLIUCEJ
B/DIASRoRAWBKowsGiShjwE6cAKoqfPjgkAUkfwcMLCAQQABaMokOGRBoAhI
d5o+HUAAgYItgSYI/NDIzVYCY8AUyGJljxSBHRil4YrgS4EGDqY80XNFoAYi
cRIouOugCoQmXHiEEUghBJA1Dx5QgRBFyZIbQd4MxMLBBxkoTpjoyGEjAp+C
Zjb0sDNkB44aESwZBMCmixYvZ8S0mZNntm+CAQEAIf4QQWRvYmUgSW1hZ2VS
ZWFkeQA7"""

Tick = """R0lGODlhEAAQAOeoACxYey9afDJjQTdrGThrGjlsHzxuHz9vIztrnDVspjtt
ljdvrkx7MUF1oUl7ZkR4r0Z9s0h+rFmFRkqAsFOMKFePLWGKSXd/gU6Ht3mA
gWaOT2mIo16Mu1iPv1SQwoSMjXGXYHGPw22Qx3SRxHGTwluZzXKWyXGZwmac
yW+bxnSds2qfxoCfz5Wen9aaDoGrYcWcLtiaCnGmzoGjzaKiooml0X6qwnWs
08akOnet04eq0oiq2Ymq2NGmOIe5YIe6YJKr1emnAMyqOtCqMou6Yo26ZIy8
ZZau1Y+7aqSusY+8Z6mxjO+uAMmyTciyUZi03HS8+teyR+iwKbKzs5i32bS0
tPS1AIq/3q25vqO63I7C3pbA26C93p7Bv4jE7Kq93prE357D36vFqJLH4q6/
36nGrqbC4LbAxb2/wpTL6afJ4rrGyejDYbLSlZ3Q57HTkOvHYabP5LTUlbHW
krbUnPHJXanR5vHLXPTLW8LP0vLPafPQc/LRfNPT0/DTgfbWfMLa8MHb8Nvb
28/f79Dmutni8dbj8dzj8drl5/bjq8zs+vnotd7t8dPw+vrsv+Pw8unw8OPy
+u7w8OLz+uvx9Ozx8+n2+un2/eb3//P19fP1+vP2+u74/fL4+O76/+v7/+j8
//X7/uv+//L9//L+//P+//f///n/////////////////////////////////
////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
/////////////////////yH+EUNyZWF0ZWQgd2l0aCBHSU1QACH5BAEKAP8A
LAAAAAAQABAAAAj8AP/9s0JQoMGBVg4ywePoThMoEKE40fOHicE6i+AI8ZLG
jZs0Wob42RNEIB82OMbYMXMECJc4XaIkkiIwBowrasgcqkSpUJYwS3q4EKgi
B5gvqCQhMmWq1JMtNlYILHGDiqFQmVpA6oQFkI4bJQR6kFHDEqdTGZJcWNPI
hAwPAjGgYDHp0qMPF/J8UiQCBQaBETrMCIQp0hlGoMqA0CDBgcAGE1LsEOVp
FCkxL+QQQlJhgEAFEE7wGLRJk4U3c1AZUeJZ4IYHHEiMCMGADqrbbVr/64MG
QYIFCwoQ+YHKB2uBVQRNCQCguQAKRdoUodCaBo2DAgcQGGDgwL+AADs="""

Plot = """R0lGODlhHwATAKIAAP///7+/v39/fz8/P////wAAAAAAAAAAACH5BAAMAP8A
LAAAAAAfABMAAANVWLrc/gzISau9QOHNpbaBII4kGVgfJQxlK7JoYYmdRFfp
dHf7lNuCGqDnkVWIGORvGKwpjaomT+qD6qib5wyb5C6Rl2cL5mSVlkIOOo1Z
s2OuuPycAAA7"""

class MenuPane(Frame) :
    def __init__(Self, Parent, relief = FLAT) :
        Frame.__init__(Self, Parent, relief = FLAT, bd=0)
        Self.Build_Menu(Parent)

    def Build_Menu(Self, Parent) :
        Self.IconImages = []
        for Image, Text, Command in ((Load, " Load ", lambda: Parent.Load()),
                                     (Save, " Save ", lambda: Parent.Save()),
                                     (Options, "Options", lambda: Parent.Options()),
                                     (Trim, " Trim ", lambda: Parent.Trim()),
                                     (Grow, " Grow ", lambda : Parent.Grow()),
                                     (RunStop, "Run/Stop", lambda: Parent.RunStop()),
                                     (Tick, " Tick ", lambda: Parent.Tick()),
                                     (Plot, " Plot ", lambda: Parent.Plot()),
                                     (Exit, " Exit ", lambda: Parent.Exit())) :
            if Self.master.ButtonMode == "name" :
                Button(Self, bd = 2, text = Text, compound = LEFT, relief = GROOVE, overrelief = RAISED, command = Command).pack(side = LEFT)
            elif Self.master.ButtonMode == "icon+name" or Self.master.ButtonMode == "icon":
                IconImage = PhotoImage(data = Image)
                Self.IconImages.append(IconImage) # keep images to avoid garbage collection
                if Self.master.ButtonMode == "icon+name" :
                     Button(Self, bd = 2, text = Text, image = IconImage, compound = LEFT, relief = GROOVE, overrelief= RAISED, command = Command).pack(side = LEFT)
                else :
                     Button(Self, bd = 2, image = IconImage, width=50, compound = LEFT, relief = GROOVE, overrelief= RAISED, command = Command).pack(side = LEFT)
            if Image == RunStop :
                Self.RunStatus = RunStatusDisplay(Self, width = 50, height = 17)
                Self.RunStatus.pack(side = LEFT)
        Self.Indicators = {}
        for Name, Color, Text in (("BadInput", "red", 'I'), ("Short", "green", 'S')) :
            Self.Indicators[Name] = Indicator(Self, Color, Text, width = 20, height = 20)
            Self.Indicators[Name].pack(side = LEFT)
        Self.Legend = Legend(Self, width = 72, height = 20, bg="white")
        Self.Legend.pack(side = LEFT)

    def Clear_Indicators(Self) :
        """ This routine clears all indicators (like at the begining of
        a simulation cycle. """
        for Name in Self.Indicators.keys() :
            Self.Indicators[Name].Update(False)

    def Destroy_Menu(Self) :
        for Child in Self.children.values() :
            Child.destroy()

    def LabelSelect(Self) :
        LabelList = Dialog(master)
        LabelList.pack(fill=X, padx=5, pady=5)

    def notdone(Self) :
        showerror("Not implemented", "Not yet available")

    def About(Self):
        showinfo("About CompuCanvas", 
                 "CompuCanvas version %1.2f\nAugust 2011\n\n"
                 "A computation exploration environment\n"
                 "Copyright (c) 2009-2011 Scott & Linda Wills\n"
                 "GUI icons thanks to Mark James at www.famfamfam.com"  % (Self.master.Version))

class RunStatusDisplay (Canvas) :
    def __init__(Self, Parent, **kwargs) :
        Canvas.__init__(Self, Parent, **kwargs)
        Self.Step = 0
        Self.CarImage = PhotoImage(data = Car)
        Self.StopImage = PhotoImage(data = Stop)
        Self.Update(False)

    def Update(Self, Run) :
        """ This routine updates the run status display. Depending on
        the run staus, run cars are displayed or a stop sign. """
        Self.delete("runstatus")
        if Run :
            Self.create_image(Self.Step * 16 + 2, 1, image=Self.CarImage, anchor=NW, tag="runstatus")
            Self.Step = (Self.Step + 1) % 3
        else :
            Self.create_image(25, 2, image=Self.StopImage, anchor=N, tag="runstatus")
            Self.Step = 0

class Indicator (Canvas) :
    def __init__(Self, Parent, Color, Text, **kwargs) :
        Canvas.__init__(Self, Parent, **kwargs)
        Self.Color = Color
        Self.Text = Text
        Self.ID = "Indicator%s%s" % (Color, Text)
        Self.Status = None
        Self.Update(False)

    def Update(Self, Status) :
        """ This routine updates the indicator with the specified status. """
        if Status <> Self.Status :
            Self.Status = Status
            Self.delete(Self.ID)
            if Status :
                Self.create_oval(2, 2, 20, 20, width=2, fill=Self.Color, tag=Self.ID)
                Self.create_text(11, 11, text=Self.Text, font=("times", 14, "bold"), anchor=CENTER, tag=Self.ID)
            else :
                Self.create_oval(2, 2, 20, 20, width=2, tag=Self.ID)

class Legend (Canvas) :
    def __init__(Self, Parent, **kwargs) :
        Canvas.__init__(Self, Parent, **kwargs)
        Self.Parent = Parent
        States = ('0', '1', 'U', 'Z', 'S')
        for I in xrange(5) :
            Text = States[I]
            Color = Self.Parent.master.Canvas.Colors[I]
            Self.create_text(2+I*14, 0, text=Text, fill=Color, font=("helvetica", 14, "normal"), anchor=NW)

######################### Proto Pane #########################

AND_Icon = """R0lGODlhQAAgAOcAAP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACwAAAAAQAAgAAAIlgABCBxIsKDBgwgTKlzIsKHD
hxAjSpxIsaLFixgzQgzAsaPHjx81SgRJsmQAkQ85jlSJkiHLlS0FmnwJs+XM
kxVpxiSoM2LPnQB+psQJlCdRikJtHp14E2VSh01FPm04VWPVhVcxZkW49WJX
g19dzrQYFutYpEuLBk0LlW3RsB3VHrwZVW5BuiXt6t3Lt6/fv4ADCxYYEAA7"""
        
OR_Icon = """R0lGODlhQAAgAOcAAP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACwAAAAAQAAgAAAIlwABCBxIsKDBgwgTKlzIsKHD
hxAjSpxIsaLFiwEyatwY4KJHgxxDavxIEkBGkCJLWjyJMKRKiiwVbnxJUKRN
hyNpmrSZEmdHnQ1jBv0JVCZRn0WNShSatOZRnzyLMh0aFehUhlebZk24NWlX
lDR59hyqUyxHqE2VLpyZtuVTp2zbHpx6U+5cs1/tmrXLt6/fv4ADCx5MMyAA
Ow=="""

XOR_Icon = """R0lGODlhQAAgAOcAAP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACwAAAAAQAAgAAAIrgABCBxIsKDBgwgTKlzIsKHD
hxAjSpxIsaLFixgzLgwQYCDHjyA7arzIUWDIkyBHTixpUiRBlCojsgQw82DI
mA5n1kSYEmdLlC53JvzokyZQnS4ZCi3a0mNSpU+ZGnUqM2pRpBKX4sRa9ahK
rhCPaqUI9uHYr0GtKjyLlqpZtRbFpn3rU6xbqHDrzt1IVGrBsgZv+v279+XJ
wTZFyu2JmLBToI0jS55MubLlyxUDAgA7"""

NOT_Icon = """R0lGODlhQAAgAOcAAP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACwAAAAAQAAgAAAImQABCBxIsKDBgwgTKlzIsKHD
hxAjSpxIsaLFixgzRgygsWPCACA5ehwpMGRIkh5BljyJEqPKgSZbWnxJMKZM
iTQL2rzpMKfBnTwV+jwINOhPkQyLUjTJtGlTiE6jRq0ptSrShlazXhyKUKlR
AFx1svx6VOhYsmWJnkWblmpYtjCveoUrduVbum7X4v2od69Qv4ADCx5MuLDE
gAA7"""

BAR_Icon = """R0lGODlhQAAgAOcAAP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACwAAAAAQAAgAAAIewABCBxIsKDBgwgTKlzIsKHD
hxAjSpxIsaLFixgzatzIUWGAAB1DDvwoUiTJkh1PotyocmXGli4vwoxZcSbN
iTZvRsy58qPPn0CDCh0KFCPRo0iP6izIc2nDpk4XQo2acCrVg1avMgWpFWLW
rgK/ghXblSzYs2jTqsUYEAA7"""

FET_Icon = """R0lGODlhQABAAPcAAP///wAAAOnp6RMTE6+vr0xMTODg4BwcHP7+/iMjIzs7
Oz09PUFBQUhISElJSRQUFB4eHgMDA4aGhoiIiJaWlqKiohAQELu7u8bGxsfH
x1BQUOjo6BsbG+vr6+3t7fDw8PT09AgICIODgwICAgAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACwAAAAAQABAAAAI6gABCBxIsKDBgwgTKlzIsKHD
hxAjSpxIsaLFixgzatzIsaPHjyBDiuyIQGDJkRkDBECpUSXLlCtfXnQp0yLN
mhRvfjxJEIHKn0CDCtX5MUTMgUOTDuX5ccBRgURxJnRaMKrUg1QJWr1aMCvS
p1yxgt0aVqBXqGDLEjwLgGxZtm7Dwk2r1uxYunXn1j1o4a5UpUKr4hU5orDh
w4aJxuX6wO/ero4fD9QreXLkypQrA8iM+bJkzp89PwY9WnRDph8DRBAMGLBI
1AB8tnatGeHix7f35q67W23vsr9rCx9OvLjx48iTK1/OvHndgAA7"""

PASS_Icon = """R0lGODlhQABAAOcAAP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACwAAAAAQABAAAAI/wABCBxIsKDBgwgTKlzIsKHD
hxAjHgwQQKLFiwUpYtxoUSPHjw49ghyZUCTJkwRNokSpciXJli5BwozJcSZN
ABQrhsxp02DOjTx1KvyJs6fAoEAr8iwpdKDNpUYXilzqE2FTp0SLJs2YVevE
plSxbvVJdKbHsCmvRnyK9CtarmOZmu36Ne7Qrm+tqoUYFWvQvXrtMvzLF/DO
h3/7HjXcsC/avHUxGoVcdDLjwYYpp2UreLFiz3s/R/Z7We5V0WRJfwyLGq7m
jj9bb5bd+DXpxKVX46arl/fIthc12l4bOzdT1TWz0l7Mlfbb5V6bGwc9GrZp
zJwlX6ZsG7pjvL5Ta2FHnJh8Z6nhj4/fSfjwevTClaZPe362WOp361c+zT97
8NC+wQSZd/1lZqBKBCKn3nXMjTdcdHfFl9R8cLEHHYS1EXfTghvq1uFv032Y
n4jJkViiie+hKNGFJLKo4oswFhQQADs="""

VDD_Icon = """R0lGODlhQAAgAPcAAP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACwAAAAAQAAgAAAIagABCBxIsKDBgwgTKlzIsKHD
hxAjSpxIsaLFixgzatzIsaPHjyBDihxJsqTDAChTqlzJsqVLlRhfypwp02RB
lDYv4sxZcSfPiT5/Rgwq9CHRog2PIl2odGnCpk6jSp1KtarVq1izat0KMiAA
Ow=="""

GND_Icon = """R0lGODlhQAAwAPcAAP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACwAAAAAQAAwAAAIhAABCBxIsKDBgwgTKlzIsKHD
hxAjSpxIsaLFixgzatzIsaNHhwECfPwYcqTHkiY5okypcSVLjC5fWowpkyLN
mhJv4oSok2XIn0CDCh1KNCjMokiTFt3JtKnTp1CjSp1KtapVpUSfYh1qtavX
r2DDih1LtuzCnl7Rml3Ltq3bt3Djyl0YEAA7"""

DOT_Icon = """R0lGODlhQAAgAPcAAP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACwAAAAAQAAgAAAIYAABCBxIsKDBgwgTKlzIsKHD
hxAjSpxIsaLFixgzatzIsaPHjyBDihxJsqTHACgDmKyYMuVKiS1bvoQY0+VM
hzVR3sSZcydPmT5/qgxKtKjRo0iTKl3KtKnTp1CjSrUYEAA7"""

LABEL_Icon = """R0lGODlhQAAgAOcAAP///wAAAPDw8GhoaKenp01NTbKysr29vZqamtDQ0MfH
x+Hh4Xx8fOnp6dnZ2YyMjAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACwAAAAAQAAgAAAI1QABCBxIsKDBgwgTKlzIsKHD
hxAjSpxIsaLFiwoNBNhoAKNHhQU2BijwsWTBBCI3JjDJEgHHjQhYlhSwcQCA
ARsFyPR4YOMBAAp87sSIM4BOmgFsDq3oAObABxsdLKXoMoBUgSgDxJwaESlJ
giGNcoXYM0BHghoD/BzrsKhOgkiVsl24IKXdlAvmLiRwty8BvQr7CgaMMKjZ
hHwDKCBs0G3CBjUZE6wbgMFCBhvzSgaQeKXCrH83b/y6MOxmw6H3blzMGHOA
Bg0hV95Mu7bt27hz6/YYEAA7"""

WIRE_Icon = """R0lGODlhQAAgAPcAAP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACwAAAAAQAAgAAAIcgABCBxIsKDBgwgTKlzIsKHD
hxAjSpxIsaLFixgzatzIsaPHjx0DiBxJsqRJkAtNqlyJMmSAlhxFwtwoc2bG
mjYv4sxZcSfPiT5/Rgwq9CHRog2PIk35cunQpk6NQo2adKXVkUivaqXKtavX
r2DDih0bEAA7"""

CLOCK_Icon = """R0lGODlhPABEAKIAAP///7+/vz8/P////39/fwAAAAAAAAAAACH5BAAMAP8A
LAAAAAA8AEQAAAP/CLrc/oAQSKu9sBTMu9deKD7gaIrlqWLp6mbbKzvtvBKC
pmtEYIuB3G6om/wwwoLAyAgQdr4jJMi7UAtM6SKgiyI1Ai0DDKyKc+ERt+D9
PdkRoq4i1Gqib3njycxlZU9pcX81fFs1L3cLhmMxi1gMfjZrDIwLhZCPgi+W
gw2YhI6cmQqdAKCfoi6dpqiNNjltraqzNqykCq6Xqje4tZW4p7wqt6F7wYgq
lI/GwFmBP4qlyLS4krZKzKkS3NcAy5PSTm3Ccn3ZR9AU3OwSUeBHZiPyR2vk
HGhiCmtw+GT6+4T8eXBloL4kSxw4gQKQYBI5RRpawGHunsQKyS6yGKbxRELG
jhg5gqTwEWC7k9x4oDwZD6LLlyVNrJxJk9vImzhz6tzJkxjMiDfzwDTYs6jR
o0iTKu3YY2kuokWbOp1KtapVMQkAADs="""

DECODER_Icon = """R0lGODlhPwA3AOYAAP///4etzXtQJO3/4wAdSEdtmU5jdi0ZAM2kff//2jMA
AD8/P83j9pl/fwAOJNi/rb+/v7J/W2eRtrLa9tH2/1EoAH9/bQ49ZP//9vbj
zW2cx/bRp/b/9lyHrRkAAA4kRa2Zf9qyhZBjPDYkDuPNtmQ9DgAAKAAADnZ/
f3aZraOOf+P2/////9js/z8tGQAOMgAoUT0OAP//7OzYv7/Y7BktP6OOjtr2
/w4OJNi2kX9/f3KavwAAM/b//3+Zrdr//9GndlqFsgAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAAMAP8A
LAAAAAA/ADcAAAf/gACCg4SFOkKIiYqLjI2KhZCRkoSHOpaXmJmamzoLQpOg
oYehpJCjpaiCp6mkq6yin6+wsq2xtJKut6a2uoa8vZS/wAC5w8TCwMXDyr3M
zcjPxr7Sg8661rTY2dDX3N3Uqt6FEJyanuDH5ZkQno6N6Ifuiwup2q/29+Ky
+Kz8qP7/9OVDd0xQhEUGGGwjeOrgog8r9gkcCOCgBkEkRggJIgiDCh0+CoGw
pDAcw1gWBwERcgFABo2IarRwCVPIiQAm4aEUclHQBiEwegi4CQCBkAIARAiR
ACAHjhQ5wTXkOegngQExHKDQYcEDDA4VTFDYddIgVUEhWCZQsOhqDB4/jMjq
NNszg4uNGCq8uFGoxIkJADDYiFhwbkVGHQTNOKBoB01FiQtLjfUgU0lBMhqQ
JDRSB41qE/uFDlh2cmlqAEulnmUY9WjSpltL40dO3aVzk21fYiePkc7eiejB
dn169mvVx1nHJi7b2Grlxos7Tw59uvRl1EE9n7Sde3Zc38E3t76cuXnXwNP7
RhUIADs="""

MUX_Icon = """R0lGODlhQQBJAOYAAP///46jtj1BPfbjxAAOPT1kjo5kPdithiQOAH92Y73j
/yRFXD8/P62GWYWy2uPEmuP//0UkDgAoUQAAGVt/sr2OZP//7K2/2L++nChR
dv/s2G1HGfbRp3+Oow4AAC0ZAAAzPWSOvZl/f6PP7FktGfb//w49ZNi/rRkZ
GZltRz0OAL+/vz9ZbX9/f0htnK2Zf8abbWSOoQAADg4kNgAZR+O9ji1Zhv//
2jMAAIePih1Iba3Y7HZRKDYkDuPNtmM8GQAAKBktWf/////sxhsAAAAzW7bN
43KaxP/asqd2UVEoAJC74+z//1F2pxktP1mGrf//9gAOJFoyDpqHdtqyf///
4//jvcTj9rLa9qfR9jNbf8Sacu3HnCRFcpFnNgAdSAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAAMAP8A
LAAAAABBAEkAAAf/gACCg4SFAGCIiYqLjI2LLYaRkpOHDC2XmJmam5yXYJCU
oZOfoqWFpKaph6Cqoqitrqywo7Kzkq+2kbi5hru8g76/gsG/xMW1wqvJvcjC
xrzPttHSzcfLp9XQ2drXwNu507Dh4t+z4wAMjo7l5KkrnZznqfLz7K30pvix
3d78yv76QgUUaE/VQEqkqCCiwURQBURF2gGEpBCMDCyCeECUyC8hGBAqQgCo
okIAmCIWRHQoAcBHiwsATrQIIOhFCyP/OlIEo2NDBgAcwOQ4eQPHl4YwwFAA
oOGDjBEPPARpeJAWAIU6khCAUAEIBqJGkSoVdADMgggTdgyzBE+TwZ0Z/2qA
ycLDBBewRwEkXSrIC6InwNQ1ensVjJYhRGKoaIIEr1i+ALYgOtJvYmEtUJSg
AOOgcZGiWwGkGAtgAIIZH6JcGVaQsMICAAxYVGDFQxEoUsCwSOBhrAUSYJZI
7sKy6q2dLvSCkVACcUQfPcA4mTK2ARgbAKD8AAPYuK7W+cCHt0y+m3dm5c2L
35d+2Xls7SmlE/wo/qR3bTO9r6zTvrP1BPmT0zX78UcggAH25x83C4KDIEIP
HidggQMe2KCDF1KToTkRfjdhh+gBRN+I6oBoSH6agMEWii2sICBhL24Y4z0m
zmigjf/haKGOyVAYo48vAvkhjz3WiKOQMhIJoQWS1vgTCAA7"""

ARITH_Icon = """R0lGODlhPQBJANUAAP///3an0R1Ibf/jvQAZR3ZRKH9/f0cZAFF2p9Gndr3j
/7J/WygAAP//7D8/P7+/vwAOMg49ZKfR9n+Oo+zGmez//z1kjpl/f2SOvY5k
Pf/20Vk/Ldr2/1oyDqO2zf///wAADi0/Wb2OZA4AAChRdn9bM1t/soWy2tqy
hfbRpwAoUVEoAJltR22Zxv//9vb//2Q9Dv//2sbs/+P//zMAAAAOPQAAGYVa
MgAdSAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAAMAP8ALAAAAAA9AEkAAAb/
QIBwSCwaj0LHA8lsOp/MnAFKrVoB0qt2W8xyv1cvePwUk89HM3otVLPR7jc5
LgfT69w7Psfv+/+AgYJ6UDkOBoiJiouMjYyEZVN4WJJfkHaVeZlyl02dmmOf
W6Jpm2+kRqhWqkSsVK5tpmywlJO1TQ6DfbJrhA+OibROtKopfAHDvFWqIjkg
Csmhyk03OQQV0ZhHC3wmRy4rOSSR0tvdRwMjORjk2kbcOd5GKHwS7Zay8PJF
LDkQHPdAAWhwIdEGPiESTXghBEaOCIWmRRQSg4YuHNg0MMiBYKK7ihexUeBz
wiM+IQQNIlTIsJkNGUNK/BEwRFefgEP0GSmQQwVDnyEy/dBscwiYxFtFdBJp
cCCHBSJB+wxFKjDpOSLpcrR4dTSb1XhFEtTjWq5Khhw1ZpB1B8VFB3HLunqS
O7Cgh7hl99BFImxu3jrE9mohlMum4DASfxntG+VwF8d42QKGXJOyyaqT/3Ky
THWz5lOcGfMNzdkrZs92bKpezdmoI0OuGS2xlao0bdF6ac+xbQt3Zt2SgQ/m
Pck3auGjiAsJAgA7"""

LATCH_Icon = """R0lGODlhPQBJALMAAP///5zH7R1Ibf/s2P/asj8/P39/f7+/v83j9v///wAA
AAAAAAAAAAAAAAAAAAAAACH5BAAMAP8ALAAAAAA9AEkAAAT/EMhJpbo46827
MlUoVkphnGiqruz6jXD4xnQ91/CN75TO96Cf0CcEEIux40+JFDF3zyZQaglS
k9Zm9LqldaXfXPbqHIcKHoyZ3EtfDi1UGFmKn2xrdpWb18/LeiN/Mn1sgySF
ZIdTMAMoCHSJeDEEGAGRfJSWmFRPlReXRYsTnpuikl6FnwqhQ6hYmqCcYKqm
rpkwq61Lr2KxrLNatbKnuCO6wckVyMWdw8DNtL+7PKN7uW4CqcYiqx3asM6/
39vigYTc59fS6ozs7evC8KS9gvXR8vNG9+j61vv6AJqb9+9fuXfwCvLjlU6d
woAP/S10l3AiPYgW47WLSDCjwI4NLM9xrMjFjcmTHu20qKNSxYGAgGCGlMmQ
JkKbt3Aq04nDIEyfGHniE9rTY4gIADs="""

REGISTER_Icon = """R0lGODlhPQBJANUAAP///7+/vy1Ofc/s/y0ZAO3HnH9bMwAdSGOQu39/fz8/
P5zH7f/s2AAOJLbN462GWfbjzfb//1ktGQ49ZEdtmRkAAE5jdg4kNn9tWePN
pOP2/3+Zrb/Y7P/246PP7DYkDg4kRRktWXKaxAAADuzYrcTj9v///zwZABkt
P+z/////7AAoUa3N40VymgAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAAMAP8ALAAAAAA9AEkAAAb/
QIBwSBS6jsikcslsuhLFqLToUiSu2Kx2y90+p+DoN0wuj8vgM3pNVLPb0Lfc
LQfQ6+H7W4+X8td/fXCCRnGEeYZ9gYeLZI2Cj2mJh36TUQpOSJaUbZlHAV1Y
kXhVoVdmm5yFjKmqo5WqU69irZyzVLWUt4NgEKYcqKxhBZ4XDojCYMSeDSWS
yVPLLhZaH0gUz4R/0gtFHSdHKxGyucHKSN1FBkcHKeTQUtxFDARHE9mQtdLU
WfVHCPgU6fPkQsC4d9oGOrmwwVE5h8OQUMOABAUwiAkjHkmXocKRBiww5tPo
Ih2AB0hAaECW8dzGISokIGnBcqTLkkQ6HhnhISAppoU4iaA8EsIdrJbR0BWJ
iUQEQptJXxYh4dFFgwFHocbipXXrKqRev3b1umtIWTRn7TwcO2etQHhh06YV
+TasWbdP7QbC5AlvHYIuQJmam8eKKXN21cIl65dWYrF1E8ttvIcyV72W7z5W
DDZuZsieF2+VC7i0k88ATF0pFYp1qACbOeeNPRsXbbqOb/u0rbv25d6/eQMX
nnW45t3D/xDuTNy47OJ4ggAAOw=="""

TOGGLE_Icon = """R0lGODlhPQBJALMAAP///5mtvzNbfy0ZAP/tx39/fyQ2Tv///0htnFszAL+/
vx0AAD8/P//2445/fwAAACH5BAAMAP8ALAAAAAA9AEkAAAT/EMhJ5bs46837
K1UoVg9TnGiqruz6jXD4xnQ91/CN75TO96Cf0CcEEIux40+JFDF3zyZQaglS
k9Zm9LqldaXfXPbqHIcYHoyZPI0pWqgwUl5mi7lruzHPow/5UIBkfiSCeIcj
DQ5wKAFYiCIEC2kXAo9UTJKUD5Z3mHyKKQMXBimOnmCGEwkXCH2qlzisD66B
kDWztTiEbbKtr7c0ucCfPMO2xTvHu7CouL9QmzbNyzYmjNPG0MzBMdVezSNh
37Fa1NvZyb60xKl6dervE7zz4UX0VfKF3e/4e/q9zAGsx0+PP3/gCto5aO8P
DTSUGi4R9IYRwnJzJLYTOPBfPH0MRjt6dNcxZEmNBE8qZGNyYEuABzfJjHiF
0YkSFq/BUSByJLyeGPcBDRpwqAxAFz8KNfqzKdOiS59Cnfr0SVKS4lAuRKoV
QAQAOw=="""

COUNTER_Icon = """R0lGODlhPQBVAOYAAP///3an0T8/P9H2/0dtmdGndgAADn9/f1VQJCQkDvbR
p///46fR9iQOAKd2USRQewAZR62/2IZZLdHjxGSOvQAAKNi/rfb//8aZbVAk
Dv/sxlEoAFt/sg4AALvj9j1kjuP2/469421HGQAoUX9/jr+/vz0OAOO9jvba
sgAOPQAOJI5kPf//9nZRKP/jvVktGeP////s2A4kRdqyhVB7ow4yWq2GWYat
2FmGrf///9r//22ZxuPEmgAAM73j/xkAAP//7AAAGf/asloyDn9/mez//yhR
dv/sz8Tj9lqFsm1IHcSacpC745BjPAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAAMAP8A
LAAAAAA9AFUAAAf/gACCg4SFhoeCAiWIjI2Oj4xOB5CUlZYAkpeam4WZnJ+X
nqCjj6Kkp4emqKuCqqyorq+ksbKgtLWct7hOvL2+v8DBwrqQTgIHyMnKy8zN
zMSlk7iY0p/QttW52bLXjd3ao9+b4qnbr+SG6JbqhOyU7q3mrPDU0/WNAsO9
8qv6vSXOktFzZCwgsmj26A30xq/SwkgN30WcNRHhNIUVsYXj18LJiAtCeuEw
pIQXh3IbDy0w4eQDgJC8VCApVNLJyXQZGR5C0cHJjpe+HtA0iVKjoQK8GAD1
xYRQzZudckI8tMJJChhLe8kAMehpKn9OLBJiMcSJEUEwfY0U5DXdMYNi/wdp
+OEkCVpePRDwCuKBLVGcKQud4BXirpMeExrwEgqgbdTAhBw4qTDAcA8dNnrd
aPz3sVFCIpzUuGBZB5AMvLY6bicVkaqVTggMgnkZAI+eTmisHoSxkAJeAWbj
1SFIwt4EnVlDFoTBiQEfwg8TB3BEsS+oyj8LqgqhSPTagpb8ws67dVFBLDaY
JUR7OgAgL66fBzfIRU8K7IcTut2LfLzlMySVn3SFNNHfV/7EZcEBJHA1SAzI
EOEdIRAiE0EqbwUUFzfmAXZRh57tAmJ2Ii7HoYnnjFheQir+9yGK87R4T4n4
gCUjQWABZNBDrmXozIYpwtiPjDzON86NOml3okeSQTIZo5CwEInkVE4OGQ5Y
WPojo0HIFBSQlwEtwmKS9hAUUZFHklkmla5Nac2ZbtLX5ppmqkmnh3Pemeee
eoZoZJ8u8rlJIAA7"""

DISPLAY_Icon = """R0lGODlhMQA9ALMAAP///7+/v+EAAH9/f/B/f////z8/P+g/P/e/vwAAAAAA
AAAAAAAAAAAAAAAAAAAAACH5BAAMAP8ALAAAAAAxAD0AAAT/EMhJZbo46817
qiCQGENpnmiqpl9IJYMrz2E7w3RO2zKu/yCeywcsinJE409YiymBzKDzqYtW
khSCYMvtegk95LSi9Zq54KHYVT6b0zWDBzPOus/wYGBlSrbzNH9XaxOCP4YT
Vi9jiDmNRzuMaECPiomSW4AzlYQSj5uTL50AnzKckXaZlKGJo6VsrBausYG0
lhaYApqmtnJzfrSgqi97fAPAwzqnN7m7sMmyqIXBvNCQzKkCBwTc3d7fB7aj
2XfU0dghbeXWotIg6uvO1z113dvf+PkIg+4SXavWbomo8++QuH4ACirr9QuL
Qke2ivFxyAWgLn7oJjysFXDcRmEXb9tl9FfRYEeEH6uFbIWy5MKTIxO6hAiT
XoWUz1aes0kBZ7qDMX2+A8pT40yOOkX4mlOPW7h1XO4R2EfM2LE606CyYxkz
q1Z5ArF4hQp2HJUwCM+qSas2iNm2GIvCrfF27s4hWO3ebaKX7o6GgH9FAAA7"""

class ProtoPane(Frame) :
    def __init__(Self, Parent, relief = FLAT) :
        Frame.__init__(Self, Parent, relief = FLAT, bd=0)
        Self.Modes = {"Switch" : ((FET_Icon, FET), (PASS_Icon, PASS), (VDD_Icon, VDD),
                                  (GND_Icon, GND), (DOT_Icon, DOT), (LABEL_Icon, LABEL),
                                  (WIRE_Icon, WIRE)),
                      "Gate" : ((AND_Icon, AND), (OR_Icon, OR),(XOR_Icon, XOR),
                                  (NOT_Icon, NOT), (BAR_Icon, BAR), (VDD_Icon, VDD),
                                  (GND_Icon, GND), (DOT_Icon, DOT), (LABEL_Icon, LABEL),
                                  (WIRE_Icon, WIRE)),
                      "Block" : ((DECODER_Icon, DECODER), (MUX_Icon, MUX), (ARITH_Icon, ARITH),
                                 (VDD_Icon, VDD), (GND_Icon, GND), (DOT_Icon, DOT), (LABEL_Icon, LABEL),
                                 (WIRE_Icon, WIRE)),
                      "State" : ((LATCH_Icon, LATCH), (REGISTER_Icon, REGISTER), (TOGGLE_Icon, TOGGLE),
                                 (DISPLAY_Icon, DISPLAY), (COUNTER_Icon, COUNTER),
                                 (CLOCK_Icon, CLOCK), (VDD_Icon, VDD), (GND_Icon, GND), (DOT_Icon, DOT),
                                 (LABEL_Icon, LABEL), (WIRE_Icon, WIRE))}
        Self.SelectedButton = None
        Self.Mode = StringVar()
        Self.Mode.set("Gate")
        Self.Menu = ""
        Self.Parent = Parent
        Self.Build_Menu()

    def Build_Menu(Self) :
        """ This routine builds the proto pane menu. All buttons display
        an image key face. A dictionary is also created to map button
        widgets to prototype objects in place mode. It also holds on to
        button images so they don't get GCed. All button clicks are mapped
        to Button_Push to change the mode for future mouse clicks. """
        if Self.Menu <> Self.Mode.get() :
            Self.Menu = Self.Mode.get()
            Self.Destroy_Menu()
            Self.Protos = {}
            for ModeOption in ["Switch", "Gate", "Block", "State"] :
                NewButton = Radiobutton(Self, text=ModeOption, variable=Self.Mode, value=ModeOption, command=Self.Build_Menu)
                NewButton.pack(anchor=W)
                Self.Protos[NewButton] = ()
            for Icon, Type in Self.Modes[Self.Menu] :
                IconImage = PhotoImage(data = Icon)
                NewButton = Button(Self, bd = 2, image = IconImage, relief = GROOVE, overrelief= RAISED)
                NewButton.pack(side = TOP)
                NewButton.bind("<Button-1>", Self.Button_Push)
                Self.Protos[NewButton] = (IconImage, Type)

    def Button_Push(Self, Event=None) :
        """ This routine handles a proto button push. """
        SelectedButton = Event.widget
        IconImage, ObjectType = Self.Protos[SelectedButton]
        Self.master.Mode = Self.master.Mode.Update(SelectedButton, ObjectType)

    def Destroy_Menu(Self) :
        """ This function removes all proto menu widgets. It may be used if the
        proto pane must be removed for screen space. """
        for Child in Self.children.values() :
            Child.destroy()

######################### Canvas View #########################

class CanvasView(Canvas) :
    def __init__(Self, *args, **kwargs) :
        Canvas.__init__(Self, *args, **kwargs)
        Self.Colors = ["blue", "red", "gray", "purple", "green", "black"]
        Self.Update_Fonts(Self.master.FontSize, Self.master.Scale)
        Self.DragBox = None
        Self.Draw_Frame()

    def Update_Fonts(Self, FontSize, Scale) :
        """ This routine updates the primary canvas display fonts. If the
        specified font size in an integer, the fants are set to that size.
        In "auto" mode, the fonts size is set from the scale factor. """
        if FontSize == "auto" :
            FontSize = [12, 14, 16, 18][Scale - 1]
        Self.Font = tkFont.Font(family="Helvetica", size=FontSize, weight="normal")
        Self.Bold = tkFont.Font(family="Helvetica", size=FontSize, weight="bold")

    def Adjust_Position(Self, WinX, WinY) :
        """ This routine adjusts a position from window coordinates to canvas coordinates. """
        CanvasX = int(Self.canvasx(WinX))
        CanvasY = int(Self.canvasy(WinY))
        return (CanvasX, CanvasY)

    def Find_Clicked_Object(Self, X, Y) :
        """ This routine finds and returns the clicked object (as a singleton). If
        no object is clicked, an empty singleton is returned. """
        ID = Self.find_closest(X, Y)
        for Tag in Self.itemcget(ID, "tag").split(' ') :
            if Tag not in ("current", "grid") :
                for Object in Self.master.Objects :
                    if Object.ID == Tag :
                        (Xmin, Ymin, Xmax, Ymax) = Self.bbox(Tag)
                        if Xmin <= X <= Xmax and Ymin <= Y <= Ymax :
                            return ( Object, )
        return ()

    def Find_Region_Objects(Self, X1, Y1, X2, Y2) :
        """ This routine finds and returns a list of objects enclosed in the specified region.
        If no object is clicked, an empty list is returned. """
        if X1 > X2 :
            X1, X2 = X2, X1
        if Y1 > Y2 :
            Y1, Y2 = Y2, Y1
        IDs = Self.find_enclosed(X1, Y1, X2, Y2)
        Objects = []
        for ID in IDs :
            for Tag in Self.itemcget(ID, "tag").split(' ') :
                if Tag not in ("current", "grid") :
                    for Object in Self.master.Objects :
                        if Object.ID == Tag and Object not in Objects :
                            Objects.append(Object)
        return Objects

    def Clear_DragBox(Self) :
        """ This routine draws a dragbox used in selection. """
        Self.master.Canvas.delete("dragbox")
        Self.DragBox = None

    def Draw_DragBox(Self, X1, Y1, X2, Y2) :
        """ This routine draws a dragbox used in selection. """
        Self.DragBox = Self.master.Canvas.create_rectangle(X1, Y1, X2, Y2, outline="blue", dash=(3,3), tag="dragbox")

    def Refresh(Self, Event=None) :
        """ This routine redraws all objects on the canvas. """
        Self.delete("all")
        Self.Draw_Frame()
        if Self.master.Grid :
            Self.Draw_Grid()
        for Object in Self.master.Objects :
            Object.Draw()
        for Object in Self.master.Selected :
            Object.Highlight(True)

    def Resize(Self) :
        """ This routine resizes the canvas based on the width and height parameters.
        The scroll bar is also adjusted. The canvas is then refreshed. """
        Width = Self.master.Width * Self.master.Scale
        Height = Self.master.Height * Self.master.Scale
        Self.configure(width=Width, height=Height)
        Self.config(scrollregion=(0, 0, Width, Height))
        Self.Refresh()

    def Draw_Frame(Self, Mode=True) :
        """ This routine draws (or undraws) a frame around the worksheet. """
        if Mode :
            Self.Create_Scaled_Rectangle(0, 0, Self.master.Width, Self.master.Height, tag="frame", outline='cyan')
        else :
            Self.delete("frame")

    def Draw_Grid(Self) :
        """ This routine draws the grid. It moves the grid under displayed objects.
        The grid flag is set. """
        for Y in range(10, Self.master.Height, 10) :
            for X in range(10, Self.master.Width, 10) :
                Self.Create_Scaled_Oval(X-1, Y-1, X+1, Y+1, fill="gray", width=0, tag="grid")
        Self.master.Grid = True

    def Erase_Grid(Self) :
        """ This routine removes the grid. The grid flag is cleared. """
        Self.delete("grid")
        Self.master.Grid = False

    def Draw_Contacts(Self) :
        """ This routine identified all contact points. """
        if Self.master.Sim.Contacts :
            for (X, Y) in Self.master.Sim.Contacts :
                Self.Create_Scaled_Line(X-3, Y-3, X+3, Y+3, fill="cyan", width=2, tag="contact")
                Self.Create_Scaled_Line(X-3, Y+3, X+3, Y-3, fill="cyan", width=2, tag="contact")

    def Erase_Contacts(Self) :
        """ This routine removes contact point marking. """
        Self.delete("contact")

    def Find_Closest_Grid(Self, X, Y) :
        """ This routine returned the closest grid position for scaled mouse position. """
        X = X / Self.master.Scale
        Extra = X % 10
        if Extra <= 5 :
            X -= Extra
        else :
            X += 10 - Extra
        Y = Y / Self.master.Scale
        Extra = Y % 10
        if Extra <= 5 :
            Y -= Extra
        else :
            Y += 10 - Extra
        return [X, Y]

    def Find_Base_Size(Self, Points, Shift=False) :
        """ This routine finds a bounding box (base and size) for
        a list of points. If the shift flag is true, all points are
        adjusted relative to the computed base. """
        MinX = MaxX = Points[0][0]
        MinY = MaxY = Points[0][1]
        for X, Y in Points :
            if X < MinX :
                MinX = X
            if X > MaxX :
                MaxX = X
            if Y < MinY :
                MinY = Y
            if Y > MaxY :
                MaxY = Y
        Base = [MinX, MinY]
        Size = (MaxX - MinX, MaxY - MinY)
        if Shift :
            for Point in Points :
                for I in (0, 1) :
                    Point[I] -= Base[I]
        return Base, Size

    def PointOnLine(Self, Point, LineEndA, LineEndB) :
        """ This function determines whether a point falls on a line segment. """
        Xp, Yp = Point
        Xa, Ya = LineEndA
        Xb, Yb = LineEndB
        if not (min(Xa, Xb) <= Xp <= max(Xa, Xb) and min(Ya, Yb) <= Yp <= max(Ya, Yb)) :
            return False
        elif Xa == Xb or Ya == Yb :
            return True
        else :
            return (Yb - Ya / Xb - Xa) == (Yp - Ya / Xp - Xa)

    def Scaled_Move(Self, ID, DeltaX, DeltaY) :
        """ This routine moves an object by a specified delta in the scaled canvas. """
        Scale = Self.master.Scale
        Self.move(ID, DeltaX*Scale, DeltaY*Scale)

    def Create_Scaled_Line(Self, X1, Y1, X2, Y2, **kwargs) :
        """ This routine draws a scaled line. """
        Scale = Self.master.Scale
        if "width" not in kwargs :
            return Self.create_line(X1*Scale, Y1*Scale, X2*Scale, Y2*Scale, width=Self.master.LineWidth, **kwargs)
        else :
            return Self.create_line(X1*Scale, Y1*Scale, X2*Scale, Y2*Scale, **kwargs)

    def Create_Scaled_Oval(Self, X1, Y1, X2, Y2, **kwargs) :
        """ This routine draws a scaled oval. """
        Scale = Self.master.Scale
        if "width" not in kwargs :
            return Self.create_oval(X1*Scale, Y1*Scale, X2*Scale, Y2*Scale, width=Self.master.LineWidth, **kwargs)
        else :
            return Self.create_oval(X1*Scale, Y1*Scale, X2*Scale, Y2*Scale, **kwargs)

    def Create_Scaled_Arc(Self, X1, Y1, X2, Y2, **kwargs) :
        """ This routine draws a scaled arc. """
        Scale = Self.master.Scale
        return Self.create_arc(X1*Scale, Y1*Scale, X2*Scale, Y2*Scale, width=Self.master.LineWidth, **kwargs)

    def Create_Scaled_Rectangle(Self, X1, Y1, X2, Y2, **kwargs) :
        """ This routine draws a scaled rectangle. """
        Scale = Self.master.Scale
        if "width" not in kwargs :
            return Self.create_rectangle(X1*Scale, Y1*Scale, X2*Scale, Y2*Scale, width=Self.master.LineWidth, **kwargs)
        else :
            return Self.create_rectangle(X1*Scale, Y1*Scale, X2*Scale, Y2*Scale, **kwargs)

    def Create_Scaled_Text(Self, X, Y, **kwargs) :
        """ This routine draws a position scaled text object. """
        Scale = Self.master.Scale
        return Self.create_text(X*Scale, Y*Scale, **kwargs)

